import glob
import json
import os
import shutil
import time
import traceback
from dataclasses import dataclass, field

from pathlib import Path
from typing import Union, Type, Optional, Dict, Callable

from data_to_paper.base_products.file_descriptions import (
    CreateDataFileDescriptions,
    DataFileDescriptions,
)
from data_to_paper.env import FOLDER_FOR_RUN, DEBUG_MODE, SCHOLAR_SERVER
from data_to_paper.interactive.base_app_startup import BaseStartDialog
from data_to_paper.servers.api_cost import StageToCost
from data_to_paper.utils.file_utils import clear_directory
from data_to_paper.utils.print_to_file import print_and_log, console_log_file_context
from data_to_paper.servers.llm_call import OPENAI_SERVER_CALLER, LLMServerCaller
from data_to_paper.servers.semantic_scholar import (
    SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER,
)

from data_to_paper.conversation.stage import Stage
from data_to_paper.conversation.actions_and_conversations import ActionsAndConversations
from data_to_paper.terminate.exceptions import TerminateException, ResetStepException
from data_to_paper.run_gpt_code.code_runner_wrapper import RUN_CACHE_FILEPATH
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.replacer import Replacer

from data_to_paper.base_steps.base_products_conversers import ProductsHandler
from data_to_paper.interactive.app_interactor import AppInteractor, _raise_if_reset
from data_to_paper.interactive import PanelNames, BaseApp
from data_to_paper.text.text_formatting import add_header_and_footer_lines


@dataclass
class BaseStepsRunner(ProductsHandler, AppInteractor):
    """
    A base class for running a series of steps whose Products gradually accumulate towards a high level goal.
    """

    ACTIONS_FILENAME = "conversation_actions.pkl"
    OPENAI_RESPONSES_FILENAME = "response_recordings.json"
    CROSSREF_RESPONSES_FILENAME = "crossref_responses.bin"
    SEMANTIC_SCHOLAR_RESPONSES_FILENAME = "semantic_scholar_responses.bin"
    SEMANTIC_SCHOLAR_EMBEDDING_RESPONSES_FILENAME = (
        "semantic_scholar_embedding_responses.bin"
    )
    CODE_RUNNER_CACHE_FILENAME = "code_runner_cache.pkl"
    API_USAGE_COST_FILENAME = "api_usage_cost.json"

    PROJECT_PARAMETERS_FILENAME = "data-to-paper.json"
    DEFAULT_PROJECT_PARAMETERS = dict()

    APP_STARTUP_CLS = BaseStartDialog

    name = None
    project_parameters: dict = field(default_factory=DEFAULT_PROJECT_PARAMETERS.copy)
    project_directory: Path = None
    temp_folder_to_run_in: Path = FOLDER_FOR_RUN
    actions_and_conversations: ActionsAndConversations = field(
        default_factory=ActionsAndConversations
    )
    should_remove_temp_folder: bool = not DEBUG_MODE

    stages_to_conversations_lens: Dict[Stage, int] = field(default_factory=dict)

    app: Optional[BaseApp] = None
    cast = None  # Type[Agent]
    should_mock: Union[bool, str] = True

    stages: Type[Stage] = Stage
    current_stage: Stage = None

    _stages_to_api_usage_cost: StageToCost = field(default_factory=StageToCost)
    stages_to_funcs: Dict[Stage, Callable] = None

    _current_exception: Optional[Exception] = None
    _prior_stage: Optional[Stage] = None

    server_caller: LLMServerCaller = None

    close_or_continue_message = dedent_triple_quote_str(
        """
        You can now:

        1. **CLOSE** the app to terminate the run. 

        2. **RE-TRY** by click the reset button of prior stages.
        """
    )

    failure_message = dedent_triple_quote_str(
        """
        ## Run Terminated
        Run terminated prematurely during the **{prior_stage}** stage.

        {exception}

        {close_or_continue_message}
        """
    )
    unexpected_error_message = dedent_triple_quote_str(
        """
        ## Run failed unexpectedly
        *data-to-paper* failed due to an unexpected error.

        ---
        ```error
        {exception}
        ```
        ---

        Please report the exception traceback from the console as a GitHub issue.

        {close_or_continue_message}        
        """
    )
    success_message = dedent_triple_quote_str(
        """
        ## Completed
        This *data-to-paper* research cycle is now completed.
        The manuscript is ready. 

        The created manuscript and all other output files are saved in:
        {output_directory}

        You can click the "{last_stage}" stage button to open the manuscript.

        Please check the created manuscript rigorously and carefully.


        **Remember that the process is not error-free and the responsibility for the final manuscript \t
        remains with you.**

        {close_or_continue_message}
        """
    )

    def _get_current_stage(self):
        return self.current_stage

    def advance_stage(self, stage: Union[Stage, bool]):
        """
        Advance the stage.
        """
        self.current_stage = stage
        if isinstance(stage, Stage) or stage is True:
            self._app_advance_stage(stage=stage)
        if isinstance(stage, Stage):
            self._add_cost_to_stage(stage=stage)
            if stage not in self.stages_to_conversations_lens:
                self.stages_to_conversations_lens[stage] = len(
                    self.actions_and_conversations.conversations
                )

    def send_product_to_client(self, product_field: str, save_to_file: bool = False):
        """
        Get the base GPT script file.
        """
        if save_to_file:
            filename = product_field + ".txt"
            with open(self.output_directory / filename, "w", encoding="utf-8") as file:
                file.write(self.products.get_description(product_field))
        if self.app:
            product = self.products.get_description_as_html(product_field)
            self._app_send_product_of_stage(
                stage=self.products.get_stage(product_field),
                product_text=product,
            )

    def reset_to_stage(self, stage: Stage):
        """
        Reset the current state to the given stage.
        This will delete the openai responses and the api usage cost files up to the given stage.
        """
        self.server_caller.reset_to_stage(stage)

        # delete all conversations in the actions_and_conversations of the steps after and including the step
        conversation_names = list(self.actions_and_conversations.conversations.keys())
        conversations_to_delete = conversation_names[
            self.stages_to_conversations_lens[stage] :
        ]
        for conversation in conversations_to_delete:
            del self.actions_and_conversations.conversations[conversation]

        # delete api usage cost up to the given stage:
        self._stages_to_api_usage_cost.delete_from_stage(stage)

        self._app_clear_stage_to_reset_to()

    def _pre_run_preparations(self):
        self._update_project_parameters()

    def _run_all_steps(self):
        """
        Run a sequence of steps towards the high level goal.
        stage can be a specific Stage, or True to indicate completion, or False to indicate early termination.
        """
        stage = self.stages.get_first()
        while True:
            self.advance_stage(stage)
            try:
                next_stage = self._run_stage(stage)
            except ResetStepException as e:
                if e.stage is True:
                    if stage is True:
                        print("Run completed successfully")
                    else:
                        print("Run terminated")
                    break  # Terminate the run
                print_and_log(f"Resetting to stage {e.stage.name}")
                next_stage = e.stage
                self.reset_to_stage(next_stage)
            except Exception as e:
                if not isinstance(e, TerminateException):
                    traceback.print_exc()
                self._current_exception = e
                next_stage = False  # Failure stage
            if next_stage is None:  # Default next stage
                try:
                    next_stage = stage.get_next()
                except ValueError:
                    next_stage = True  # Success stage
            self._prior_stage = stage
            stage = next_stage

    @_raise_if_reset
    def _check_for_reset(self):
        """
        Just wait and raise a ResetStepException if the user hit the reset-to-step buttons.
        """
        if not self.app:
            raise ResetStepException(True)
        time.sleep(0.1)

    def _wait_for_reset(self):
        """
        Wait for the user to hit the reset-to-step buttons.
        """
        while True:
            self._check_for_reset()

    def _failed_step(self):
        """
        Handle a failed step. Either because of an unexpected error or because of
        graceful step termination.
        """
        e = self._current_exception
        self._current_exception = None
        if isinstance(e, TerminateException):
            self._respond_to_terminate_exception(e)
        else:
            self._respond_to_unexpected_error(e)
        self._wait_for_reset()

    def _respond_to_terminate_exception(self, e: TerminateException):
        msg = Replacer(
            self,
            self.failure_message,
            kwargs={"exception": str(e), "prior_stage": self._prior_stage.value},
        ).format_text()
        self._app_send_prompt(PanelNames.MISSION_PROMPT, msg, from_md=True)
        self._app_set_header("Terminate upon failure")
        print_and_log(add_header_and_footer_lines("TERMINATING RUN", msg))

    def _respond_to_unexpected_error(self, e: Exception):
        self._app_clear_panels()
        msg = Replacer(
            self, self.unexpected_error_message, kwargs={"exception": str(e)}
        ).format_text()
        self._app_send_prompt(PanelNames.MISSION_PROMPT, msg, from_md=True)
        self._app_set_header("Unexpected Error")
        print_and_log(add_header_and_footer_lines("UNEXPECTED ERROR", msg))

    def _finished_step(self):
        """
        Handle a finished step.
        """
        msg = Replacer(
            self,
            self.success_message,
            kwargs={"last_stage": self.stages.get_last().value},
        ).format_text()
        self._app_send_prompt(PanelNames.MISSION_PROMPT, msg, from_md=True)
        self._app_set_header("Completed")
        print_and_log(add_header_and_footer_lines("COMPLETED RUN", msg))
        self._wait_for_reset()

    def _run_stage(self, stage: Stage) -> Optional[Stage]:
        """
        Return the next stage to run.
        `None` for default next stage.
        `True` when the run is completed.
        `False` when the run is gracefully terminated.
        """
        if stage is True:
            func = self._finished_step
        elif stage is False:
            func = self._failed_step
        else:
            func = self.stages_to_funcs[stage]
        return func()

    def _get_files_to_keep(self):
        """
        Get the files to keep after the run.
        """
        return [
            str(self.output_directory / recording_file)
            for recording_file in [
                self.CODE_RUNNER_CACHE_FILENAME,
                self.OPENAI_RESPONSES_FILENAME,
                self.CROSSREF_RESPONSES_FILENAME,
                self.SEMANTIC_SCHOLAR_RESPONSES_FILENAME,
                self.SEMANTIC_SCHOLAR_EMBEDDING_RESPONSES_FILENAME,
                self.API_USAGE_COST_FILENAME,
            ]
        ]

    def _create_or_clean_output_folder(self):
        """
        Create empty output folder (delete all files if exists).
        """
        if os.path.exists(self.output_directory):
            # delete all the files except the mock_openai file:
            for file in glob.glob(str(self.output_directory / "*")):
                if file not in self._get_files_to_keep():
                    # the file can be a non-empty directory or a file. remove it anyway:
                    if os.path.isfile(file):
                        os.remove(file)
                    else:
                        shutil.rmtree(file)
        else:
            os.makedirs(self.output_directory)

    def _create_temp_folder_to_run_in(self):
        """
        Create a temporary folder to run the code in.
        """
        clear_directory(self.temp_folder_to_run_in, create_if_missing=True)

    def _get_path_in_output_directory(self, file_name: str = None):
        return self.output_directory / file_name if self.should_mock else None

    def run_all_steps(self):
        """
        Run all steps and save all created files to the output folder.
        """

        @RUN_CACHE_FILEPATH.temporary_set(
            self._get_path_in_output_directory(self.CODE_RUNNER_CACHE_FILENAME)
        )
        @SCHOLAR_SERVER.get_server_instance().record_or_replay(
            self._get_path_in_output_directory(
                self.CROSSREF_RESPONSES_FILENAME
                if SCHOLAR_SERVER.server == "Crossref"
                else self.SEMANTIC_SCHOLAR_RESPONSES_FILENAME
            )
        )
        @SEMANTIC_SCHOLAR_EMBEDDING_SERVER_CALLER.record_or_replay(
            self._get_path_in_output_directory(
                self.SEMANTIC_SCHOLAR_EMBEDDING_RESPONSES_FILENAME
            )
        )
        @OPENAI_SERVER_CALLER.record_or_replay(
            self._get_path_in_output_directory(self.OPENAI_RESPONSES_FILENAME),
            fail_if_not_all_responses_used=False,
        )
        def run():
            self._run_all_steps()

        self.server_caller = OPENAI_SERVER_CALLER
        self.server_caller.set_current_stage_callback(self._get_current_stage)
        self.server_caller.set_api_cost_callback(self._add_cost_to_stage)
        self._create_or_clean_output_folder()
        self._create_temp_folder_to_run_in()
        self._pre_run_preparations()
        with console_log_file_context(self.output_directory / "console_log.txt"):
            try:
                run()
            finally:
                self.server_caller.set_current_stage_callback()
                self.server_caller.set_api_cost_callback()
                if self.should_remove_temp_folder:
                    # remove temp folder and all its content:
                    shutil.rmtree(self.temp_folder_to_run_in, ignore_errors=True)

    @classmethod
    def get_project_parameters_from_project_directory(
        cls, project_directory: Path, add_default_parameters: bool = True
    ) -> dict:
        """
        Get the project parameters from the project directory.
        """
        if add_default_parameters:
            project_parameters = cls.DEFAULT_PROJECT_PARAMETERS.copy()
        else:
            project_parameters = {}
        if cls.PROJECT_PARAMETERS_FILENAME:
            with open(project_directory / cls.PROJECT_PARAMETERS_FILENAME) as file:
                input_project_parameters = json.load(file)
            # check that the keys of input_project_parameters are in project_parameters:
            unknown_keys = set(input_project_parameters.keys()) - set(
                cls.DEFAULT_PROJECT_PARAMETERS.keys()
            )
            if unknown_keys:
                raise ValueError(f"Unknown keys in project parameters: {unknown_keys}")
            project_parameters.update(input_project_parameters)
        return project_parameters

    @classmethod
    def create_project_directory_from_project_parameters(
        cls, project_directory: Path, project_parameters: dict, **kwargs
    ):
        """
        Create the project directory from the project parameters.
        """
        clear_directory(project_directory, create_if_missing=True)
        if cls.PROJECT_PARAMETERS_FILENAME:
            with open(project_directory / cls.PROJECT_PARAMETERS_FILENAME, "w") as file:
                json.dump(project_parameters, file, indent=4)

    @classmethod
    def check_files_exist(cls, project_directory: Path, project_parameters: dict):
        """
        Check that we have all the files needed for the run.
        raise FileNotFoundError if a file is missing.
        """
        return

    def _update_project_parameters(self):
        """
        Get the project parameters from the project directory.
        """
        self.project_parameters = self.get_project_parameters_from_project_directory(
            self.project_directory
        )

    """
    api usage cost
    """

    def _add_cost_to_stage(self, cost: float = 0, stage: Optional[Stage] = None):
        stage = stage or self.current_stage
        self._stages_to_api_usage_cost[stage] = (
            self._stages_to_api_usage_cost.get(stage, 0) + cost
        )
        self._stages_to_api_usage_cost.save_to_json(
            self.output_directory / self.API_USAGE_COST_FILENAME
        )
        self.app_send_api_usage_cost()

    def app_send_api_usage_cost(self):
        self._app_send_api_usage_cost(self._stages_to_api_usage_cost)


@dataclass
class DataStepRunner(BaseStepsRunner):
    """
    A class for running a series of steps towards a high level goal.
    With the ability to handle data files.
    """

    data_file_descriptions: DataFileDescriptions = field(
        default_factory=DataFileDescriptions
    )
    DEFAULT_PROJECT_PARAMETERS = dict(
        data_filenames=[],
        data_files_is_binary=[],
        general_description="",
        data_file_descriptions=[],
    )

    @classmethod
    def get_project_parameters_from_project_directory(
        cls, project_directory: Path, add_default_parameters: bool = True
    ) -> dict:
        """
        Get the project parameters from the project directory.
        """
        project_parameters = super().get_project_parameters_from_project_directory(
            project_directory, add_default_parameters
        )
        # add data_file descriptions:
        data_file_descriptions = CreateDataFileDescriptions(
            data_files_str_paths=project_parameters["data_filenames"],
            project_directory=project_directory,
            data_files_is_binary=[None] * len(project_parameters["data_filenames"]),
        ).get_raw_str_data_file_descriptions()
        project_parameters["data_file_descriptions"] = [
            data_file_description.description
            for data_file_description in data_file_descriptions
        ]
        project_parameters["general_description"] = (
            data_file_descriptions.general_description
        )
        return project_parameters

    @classmethod
    def create_project_directory_from_project_parameters(
        cls, project_directory: Path, project_parameters: dict, **kwargs
    ):
        """
        Create the project directory from the project parameters.
        """
        project_parameters = project_parameters.copy()
        data_file_descriptions = project_parameters.pop("data_file_descriptions")
        general_description = project_parameters.pop("general_description")
        super().create_project_directory_from_project_parameters(
            project_directory, project_parameters
        )
        CreateDataFileDescriptions(
            project_directory=project_directory,
            data_files_str_paths=project_parameters["data_filenames"],
        ).create_file_descriptions(
            general_description=general_description,
            data_file_descriptions=data_file_descriptions,
        )

    @classmethod
    def check_files_exist(cls, project_directory: Path, project_parameters: dict):
        """
        Check that the files exist.
        """
        super().check_files_exist(project_directory, project_parameters)
        CreateDataFileDescriptions(
            project_directory=project_directory,
            data_files_str_paths=project_parameters["data_filenames"],
        ).check_files_exist()

    def _read_data_file_descriptions(self):
        """
        Read the data file descriptions from the project directory
        """
        self.data_file_descriptions = CreateDataFileDescriptions(
            data_files_str_paths=self.project_parameters["data_filenames"],
            project_directory=self.project_directory,
            data_files_is_binary=self.project_parameters["data_files_is_binary"],
            temp_folder_to_run_in=self.temp_folder_to_run_in,
        ).create_temp_folder_and_get_file_descriptions()

    def _pre_run_preparations(self):
        super()._pre_run_preparations()
        self._read_data_file_descriptions()



from typing import Optional, Union, Tuple, List

import matplotlib as mpl
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pandas import DataFrame

from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.check_type import raise_on_wrong_func_argument_types_decorator
from data_to_paper.text.highlighted_text import text_to_html
from data_to_paper.utils.numerics import is_lower_eq

from data_to_paper.run_gpt_code.overrides.pvalue import PValueToStars, OnStr, OnStrPValue
from .consts import DfColumnTyping
from .describe import describe_value, describe_df
from .matplotlib_utils import get_xy_coordinates_of_df_plot, \
    replace_singleton_legend_with_axis_label, add_grid_line_at_base_if_needed, rotate_xticklabels_if_not_numeric
from ..run_gpt_code.overrides.dataframes.utils import df_to_html_with_value_format

RC_PARAMS = {
    'savefig.facecolor': 'white',
    'axes.facecolor': 'white'
}

NoneType = type(None)
ColumnChoice = Union[DfColumnTyping, NoneType, List[DfColumnTyping]]
ColumnChoiceWithPairs = \
    Union[DfColumnTyping, NoneType, List[DfColumnTyping],
          Tuple[DfColumnTyping, DfColumnTyping], List[Tuple[DfColumnTyping, DfColumnTyping]]]


example_plotting = dedent_triple_quote_str("""
    Example of proper use of the `df_to_figure`:

    # Example 1: ci as (low, high) tuples
    df = pd.DataFrame({
        'apple': [1, 2, 3],
        'banana': [4, 5, 6],
        'apple_ci': [(0.9, 1.1), (1.8, 2.2), (2.7, 3.3)],
        'banana_ci': [(3.9, 4.1), (4.8, 5.2), (5.7, 6.3)],
        'apple_p_value': [0.1, 0.05, 0.001],
        'banana_p_value': [0.1, 0.05, 0.001],
    })

    df_to_figure(df, 'example1', kind='bar', y=['apple', 'banana'], 
        y_ci=['apple_ci', 'banana_ci'],
        y_p_value=['apple_p_value', 'banana_p_value']
    )

    # Example 2: ci as two separate columns
    df = pd.DataFrame({
        'apple': [1, 2, 3],
        'banana': [4, 5, 6],
        'apple_low': [0.9, 1.8, 2.7],
        'apple_high': [1.1, 2.2, 3.3],
        'banana_low': [3.9, 4.8, 5.7],
        'banana_high': [4.1, 5.2, 6.3],
        'apple_p_value': [0.1, 0.05, 0.001],
        'banana_p_value': [0.1, 0.05, 0.001],
    })

    df_to_figure(df, 'example2', kind='bar', y=['apple', 'banana'], 
        y_ci=[('apple_low', 'apple_high'), ('banana_low', 'banana_high')], 
        y_p_value=['apple_p_value', 'banana_p_value']
    )

    """)


def _get_errors(df: pd.DataFrame, columns: ColumnChoiceWithPairs, arg_name: str, scalars_only: bool = None,
                is_list: bool = False, xy_name: str = 'y') -> Optional[np.ndarray]:
    """
    We allow referring to columns as strings or iterables of strings, or tuples of two strings, or iterables of tuples.
    Each such column is expected to contain either scalar values or tuples of two scalar values.

    scalars_only:
        True:  only columns with scalar values are allowed. Return array (n, 1, m)
        False: only tuples of two scalar values are allowed. Return array (n, 2, m)
        None:  both are allowed. scalar columns are converted to tuples of two identical values.
                  Return array of shape (n, 2, m).

    is_list:
        Designates whether the primary (namely the corresponding 'x', or 'y') argument is a list.
    """
    if columns is None:
        return None
    if isinstance(columns, (str, Tuple)):
        columns = [columns]
    results = []
    for column in columns:
        err = None
        if isinstance(column, Tuple):
            column = list(column)
        result = np.array(np.array(df[column]).tolist()).reshape(len(df), -1)
        if result.shape[1] == 1:
            assert isinstance(column, str)

            if scalars_only is False:
                err = dedent_triple_quote_str(f"""
                    either:
                        - a single string referring to a single column containing two-value tuples (ci_low, ci_high)
                        - a tuple of two strings referring to two columns containing scalar values for \t
                    the low and high values.
                    But column `{column}` contains scalar values.
                    """)
            if scalars_only is None:
                result = np.concatenate([result, result], axis=1)
        else:
            if scalars_only is True:
                err = dedent_triple_quote_str(f"""
                    a string referring to a column with scalar values.
                    But column `{column}` contains multi-dimensional values.
                    """)
        if err:
            if is_list:
                pre_msg = f"When the `{xy_name}` argument is a list of columns, " \
                          f"`{arg_name}` should be a list of matching length, with each element being "
            else:
                pre_msg = f"When the `{xy_name}` argument is a single string (referring to a column), " \
                          f"`{arg_name}` should be "
            raise ValueError(pre_msg + err + '\n\n' + example_plotting)

        results.append(result.T)
    results = np.array(results)
    return results


def _convert_err_and_ci_to_err(df: pd.DataFrame, xy: ColumnChoice, xy_name: str,
                               err: ColumnChoiceWithPairs, ci: ColumnChoiceWithPairs) -> Optional[np.ndarray]:
    """
    Create a confidence interval from a dataframe.
    """
    err_name = f'{xy_name}err'
    ci_name = f'{xy_name}_ci'
    if ci is None and err is None:
        return None
    if err is not None and ci is not None:
        raise ValueError(f'Error bars should be specified with either `{err_name}` or `{ci_name}` arguments, '
                         f'not both.\n\n{example_plotting}')
    if err is not None:
        return _get_errors(df, err, err_name, scalars_only=None, is_list=isinstance(xy, List), xy_name=xy_name)
    ci_vals = _get_errors(df, ci, ci_name, scalars_only=False, is_list=isinstance(xy, List), xy_name=xy_name)
    nominal = df[xy].to_numpy().T  # (n, m)
    errs = np.array([nominal - ci_vals[:, 0, :], ci_vals[:, 1, :] - nominal]).swapaxes(0, 1)
    if np.any(errs < 0):
        raise ValueError(f'Confidence intervals for {xy_name} values must be provided in the form of '
                         f'(low, high) tuples. These values must FLANK the nominal value, '
                         f'but the provided values do not.\n\n{example_plotting}')
    return errs


def _check_matching_column_choice(primary_name, optional_name, primary: ColumnChoice, optional: ColumnChoiceWithPairs):
    if primary is None:
        if optional is None:
            return
        raise ValueError(f'The `{optional_name}` argument cannot be provided without the `{primary_name}` argument.'
                         f'\n\n{example_plotting}')
    elif optional is None:
        pass
    elif isinstance(primary, str):
        if not isinstance(optional, (str, tuple)):
            raise ValueError(f'If `{primary_name}` is str, `{optional_name}` must be a str, or Tuple[str, str].'
                             f'\n\n{example_plotting}')
    elif isinstance(primary, List):
        if not isinstance(optional, List) or len(primary) != len(optional):
            raise ValueError(f'If `{primary_name}` is a list, `{optional_name}` must be a list with the same length.'
                             f'\n\n{example_plotting}')
    else:
        raise ValueError(f'`{primary_name}` should be a string or an list of strings.'
                         f'\n\n{example_plotting}')


def _get_custom_colors(num_colors: int):
    if num_colors < 2:
        return ["#555599"]
    elif num_colors < 9:
        return ["#555599", '#55aa55', '#994444', "#aaaa22", "#55aaaa", '#994499', "#5599ee", "#ee9955", "#99ee55"]
    return [color['color'] for color in plt.rcParams['axes.prop_cycle']]


@raise_on_wrong_func_argument_types_decorator
def df_plot_with_pvalue(df: DataFrame, x: Optional[DfColumnTyping] = None, y: ColumnChoice = None,
                        kind: str = 'bar', ax: Optional[plt.Axes] = None,
                        xerr: Optional[DfColumnTyping] = None, yerr: ColumnChoiceWithPairs = None,
                        y_ci: ColumnChoiceWithPairs = None,
                        y_p_value: ColumnChoice = None,
                        **kwargs):
    """
    Same as df.plot, but allows for plotting p-values as 'ns', '*', '**', or '***'.
    p_value: A string representing the column name of the p-values.
    """
    if y is None:
        raise ValueError('The `y` argument must be provided, indicating the column(s) to plot.')

    _check_matching_column_choice('y', 'y_p_value', y, y_p_value)
    _check_matching_column_choice('y', 'yerr', y, yerr)
    _check_matching_column_choice('y', 'y_ci', y, y_ci)
    xerr = _convert_err_and_ci_to_err(df, x, 'x', xerr, None)
    yerr = _convert_err_and_ci_to_err(df, y, 'y', yerr, y_ci)

    kwargs = kwargs.copy()
    num_y_cols = len(y) if isinstance(y, list) else 1
    kwargs['color'] = kwargs.get('color', _get_custom_colors(num_y_cols))
    kwargs['edgecolor'] = kwargs.get('edgecolor', 'black')
    kwargs['linewidth'] = kwargs.get('linewidth', 1)
    kwargs['fontsize'] = kwargs.get('fontsize', 9)
    with mpl.rc_context(rc=RC_PARAMS):
        try:
            ax = df.plot(x=x, y=y, kind=kind, ax=ax, xerr=xerr, yerr=yerr, **kwargs)
        except Exception as e:
            msg = f'Error calling df.plot(x={describe_value(x)}, y={describe_value(y)}, kind={describe_value(kind)}, ' \
                  f'xerr={describe_value(xerr)}, yerr={describe_value(yerr)}, **{describe_value(kwargs)}):\n' \
                  f'Got the following exception:\n{e}'
            raise ValueError(msg)
        coords = get_xy_coordinates_of_df_plot(df, x=x, y=y, kind=kind)
        replace_singleton_legend_with_axis_label(ax, kind)

        if y_p_value:
            y_p_values = _get_errors(df, y_p_value, 'y_p_value', scalars_only=True,
                                     is_list=isinstance(y, List), xy_name='y')
            if yerr is None:
                raise ValueError('The `y_ci` argument must be provided when including `y_p_value`.')
            min_y, max_y = ax.get_ylim()
            for col_index, index_data in coords.items():
                for row_index, (x, y) in index_data.items():
                    p_val = y_p_values[col_index, 0, row_index]
                    if not isinstance(p_val, float):
                        raise ValueError(f'P-values must be numeric, but found `{type(p_val)}`.')
                    if p_val < 0 or p_val > 1:
                        raise ValueError(f'P-values must be between 0 and 1, but found `{p_val}`.')
                    errs = yerr[col_index, :, row_index]
                    if kind == 'bar' and y < 0:
                        y_plt = y - errs[0] - (max_y - min_y) * 0.01
                        va = 'top'
                    else:
                        y_plt = y + errs[1] + (max_y - min_y) * 0.01
                        va = 'baseline'
                    text = ax.text(x, y_plt, PValueToStars(p_val).convert_to_stars(), ha='center', va=va, fontsize=7)
                    bbox = text.get_window_extent()
                    inv = ax.transData.inverted()
                    bbox_data = inv.transform_bbox(bbox)
                    text_height = bbox_data.intervaly[1] - bbox_data.intervaly[0]
                    min_y = min(min_y, min(bbox_data.intervaly) - text_height)
                    max_y = max(max_y, max(bbox_data.intervaly) + text_height)
            ax.set_ylim(min_y, max_y)
        if kind == 'bar':
            add_grid_line_at_base_if_needed(ax, 'h')
        if kind == 'barh':
            add_grid_line_at_base_if_needed(ax, 'v')
        rotate_xticklabels_if_not_numeric(ax)

    return ax


def get_description_of_plot_creation(df, fig_filename, kwargs, is_html: bool = True,
                                     should_format: bool = False,
                                     max_rows_and_columns_to_show: Tuple[Optional[int], Optional[int]] = (None, None)
                                     ) -> str:
    """
    Get a description of how the plot was created.
    This is what the LLM will get. This is essentially how the LLM "sees" the figure.
    More sophisticated implementations can be added in the future.
    """
    max_rows_to_show, max_columns_to_show = max_rows_and_columns_to_show
    if is_html:
        len_df = len(df)
        too_long = not is_lower_eq(len_df, max_rows_to_show)
        if too_long:
            df = df.head(3)
        df_str = df_to_html_with_value_format(df, border=0, justify='left')
        if too_long:
            df_str += f'<br>... total {len_df} rows\n'
    else:
        with OnStrPValue(OnStr.SMALLER_THAN):
            df_str = describe_df(df, should_format=should_format,
                                 max_rows_and_columns_to_show=max_rows_and_columns_to_show)

    kwargs = kwargs.copy()
    ci_x = kwargs.pop('x_ci', None)
    ci_y = kwargs.pop('y_ci', None)
    p_value_x = kwargs.pop('x_p_value', None)
    p_value_y = kwargs.pop('y_p_value', None)

    h = f'This latex figure presents "{fig_filename}",\nwhich was created from the df:\n\n'

    s = '\n\n'
    s += f'To create the figure, this df was plotted with the command:\n\n'
    s += f'df.plot({", ".join([f"{k}={repr(v)}" for k, v in kwargs.items()])})'
    if ci_x:
        s += f'\n\nConfidence intervals for x-values were then plotted based on column: {repr(ci_x)}.'
    if ci_y:
        s += f'\n\nConfidence intervals for y-values were then plotted based on column: {repr(ci_y)}.'
    if p_value_x:
        s += f'\n\nP-values for x-values were taken from column: {repr(p_value_x)}.'
    if p_value_y:
        s += f'\n\nP-values for y-values were taken from column: {repr(p_value_y)}.'
    if p_value_x or p_value_y:
        s += f'\n\nThese p-values were presented above the data points as stars ' \
             f'(with significance threshold values indicated in the figure caption).'
    if is_html:
        h = text_to_html(h)
        s = text_to_html(s)
    return h + df_str + s



from __future__ import annotations

import sys
from functools import partial
from pathlib import Path
from typing import List, Tuple, Type, Optional, TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication, QDialog, QVBoxLayout, QLabel, QLineEdit, QPushButton, QFileDialog, \
    QMessageBox, QWidget, QHBoxLayout, QSizePolicy, QFrame, QCheckBox, QComboBox

from data_to_paper.base_products.file_descriptions import TEXT_EXTS
from data_to_paper.env import BASE_FOLDER

from .base_widgets import PlainTextPasteTextEdit
from .get_app import get_or_create_q_application_if_app_is_pyside
from .styles import SCROLLBAR_STYLE, QCHECKBOX_STYLE

if TYPE_CHECKING:
    from data_to_paper.base_steps import BaseStepsRunner

BASE_PROJECT_DIRECTORY = BASE_FOLDER / 'projects'


def get_relative_path(base_path, target_path):
    """
    Get the relative path from base_path to target_path. Allowing for .. in the path.
    """
    base_parts = base_path.parts
    target_parts = target_path.parts
    common_length = 0
    for base_part, target_part in zip(base_parts, target_parts):
        if base_part != target_part:
            break
        common_length += 1
    back_steps = len(base_parts) - common_length
    forward_steps = target_parts[common_length:]
    return Path(*(['..'] * back_steps + list(forward_steps)))


style_sheet = """
QWidget {
    background-color: #404040; /* Application Background Color */
    color: white;
    font-family: Arial, Arial;
    font-size: 14pt;
}

QPushButton {
    background-color: #6060a0;
    border: none;
    color: #ffffff; /* White */
    padding: 10px 24px;
    text-align: center;
    text-decoration: none;
    font-size: 16px;
    margin: 4px 2px;
    border-radius: 12px;
}

QPushButton:hover {
    background-color: #7070c0;
}

QPushButton:pressed {
    background-color: #505080;
}

QPushButton:disabled {
    background-color: #606060;
}

QComboBox {
    background-color: #151515;
    color: white;
    border: 1px solid #606060;
    border-radius: 5px;
    padding: 5px;
}

QLineEdit {
    height: 25px; /* Adjust the height here */
    padding: 5px 10px;
}
""" + SCROLLBAR_STYLE

text_box_style = "background-color: #151515; color: white;"


def create_info_label(tooltip_text):
    info_label = QLabel("ℹ️")
    info_label.setToolTip(tooltip_text)
    return info_label


class FileWidget(QWidget):
    def __init__(self, abs_project_directory: Optional[Path], file_number: int, file_path='', is_binary=False,
                 description='', on_change=lambda: None):
        super().__init__()
        self._abs_project_directory = abs_project_directory
        self._abs_file_path = None

        file_input_widget = QFrame()  # Create a QFrame for encapsulating the file input widget
        file_input_widget.setFrameShape(QFrame.Box)  # noqa
        file_input_widget.setFrameShadow(QFrame.Plain)  # noqa

        layout = QVBoxLayout(file_input_widget)  # Maintain horizontal layout
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)

        file_path_tray = QHBoxLayout()

        description_info = create_info_label(
            "Add a data file by clicking the 'Browse' button.\n"
            "Provide a description of the file, including the type of data and attribute names.")
        file_path_tray.addWidget(description_info)

        file_label_widget = QLabel(f"File #{file_number}:")
        file_path_tray.addWidget(file_label_widget)
        self.file_label_widget = file_label_widget

        file_path_widget = QLineEdit()
        file_path_widget.setReadOnly(True)
        file_path_widget.setStyleSheet(text_box_style)
        file_path_tray.addWidget(file_path_widget)
        self.file_path_widget = file_path_widget

        is_binary_checkbox = QCheckBox("Binary")
        is_binary_checkbox.setStyleSheet(QCHECKBOX_STYLE)
        is_binary_checkbox.setChecked(is_binary)
        file_path_tray.addWidget(is_binary_checkbox)
        self.is_binary_checkbox = is_binary_checkbox

        browse_button = QPushButton("Browse")
        browse_button.clicked.connect(self.browse_files)
        file_path_tray.addWidget(browse_button)
        self.browse_button = browse_button

        remove_button = QPushButton("Delete")
        file_path_tray.addWidget(remove_button)
        self.remove_button = remove_button

        layout.addLayout(file_path_tray)

        description_edit = PlainTextPasteTextEdit()
        description_edit.setPlaceholderText("Enter file description here...")
        description_edit.setStyleSheet(text_box_style)
        description_edit.setPlainText(description)
        layout.addWidget(description_edit)
        self.description_edit = description_edit

        self.set_file_path(file_path)

        self.file_path_widget.textChanged.connect(on_change)
        self.description_edit.textChanged.connect(on_change)

    @property
    def abs_path(self) -> str:
        return self._abs_file_path

    @property
    def abs_or_rel_path(self) -> str:
        return self._convert_abs_file_path_to_abs_or_rel(self._abs_file_path)

    @property
    def is_binary(self) -> bool:
        return self.is_binary_checkbox.isChecked()

    @property
    def description(self) -> str:
        return self.description_edit.toPlainText()

    def setDisabled(self, disable: bool):
        self.is_binary_checkbox.setDisabled(disable)
        self.browse_button.setDisabled(disable)
        self.remove_button.setDisabled(disable)
        self.description_edit.setReadOnly(disable)

    def browse_files(self):
        search_in = str(BASE_PROJECT_DIRECTORY) if self._abs_project_directory is None \
            else str(self._abs_project_directory)
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Select a data file", search_in,
            "CSV Files (*.csv);;TEXT Files (*.txt);;Excel Files (*.xlsx, *.xls);;ZIP Files (*.zip);;All Files (*)")
        if file_path:
            file_path = file_path.replace('.zip', '')
            self.set_file_path(file_path)
            self.set_is_binary()

    def set_is_binary(self, is_binary: Optional[bool] = None):
        """
        Set the is_binary checkbox. If is_binary is None, it will be determined based on the file extension.
        """
        if is_binary is None:
            if self._abs_file_path:
                ext = self._abs_file_path.split('.')[-1]
                is_binary = '.' + ext not in TEXT_EXTS
            else:
                is_binary = False
        self.is_binary_checkbox.setChecked(is_binary)

    def set_file_path(self, file_path: str):
        """
        Set the file path. If the path is relative, it will be converted to an absolute path,
        relative to the project directory.
        """
        if not file_path:
            file_path = None
        else:
            file_path = Path(file_path)
            if not file_path.is_absolute():
                if self._abs_project_directory is None:
                    raise ValueError("Project directory is not set.")
                file_path = self._abs_project_directory / file_path
            file_path = str(file_path)
        self._abs_file_path = file_path
        self._refresh_file_path()

    def set_project_directory(self, abs_project_directory: Optional[Path]):
        self._abs_project_directory = abs_project_directory
        self._refresh_file_path()

    def _refresh_file_path(self):
        file_path = self._convert_abs_file_path_to_abs_or_rel(self._abs_file_path)
        self.file_path_widget.setText(file_path)

    def _convert_abs_file_path_to_abs_or_rel(self, abs_file_path: str) -> str:
        if not abs_file_path:
            return ''
        abs_file_path = Path(abs_file_path)
        if self._abs_project_directory is not None:
            try:
                abs_file_path = get_relative_path(self._abs_project_directory, abs_file_path)
            except ValueError:
                pass
        return str(abs_file_path)


class MultiFileWidget(QWidget):
    def __init__(self, abs_project_directory: Optional[Path] = None, file_paths: List[str] = None,
                 is_binary: List[bool] = None, descriptions: List[str] = None, on_change=lambda: None):
        file_paths = file_paths or []
        is_binary = is_binary or []
        descriptions = descriptions or []
        super().__init__()

        self.current_file_widget = None
        self.file_widgets = []

        self._abs_project_directory = abs_project_directory
        layout = QVBoxLayout()

        # Create a horizontal layout for the pulldown menu and file count label
        header_layout = QHBoxLayout()

        # Create the ComboBox (dropdown) to list files
        self.file_selector = QComboBox()
        self.file_selector.currentIndexChanged.connect(self.display_selected_file)

        # Set the size policy to make the ComboBox extend to the right as far as possible
        self.file_selector.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        header_layout.addWidget(self.file_selector)

        # Create a label to show the total number of files
        self.file_count_label = QLabel(f"")
        self.update_file_count()
        self.file_count_label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)  # Fixed size for the label
        header_layout.addWidget(self.file_count_label)

        # Add Another File button
        self.add_file_button = QPushButton("Add Another File")
        self.add_file_button.clicked.connect(self.add_file)
        self.add_file_button.setFixedWidth(200)  # prevent button from expanding
        header_layout.addWidget(self.add_file_button, alignment=Qt.AlignmentFlag.AlignRight)

        # Add the header layout to the main layout
        layout.addLayout(header_layout)

        # Area to display the selected FileWidget
        self.file_display_area = QVBoxLayout()
        layout.addLayout(self.file_display_area)

        self.setLayout(layout)
        for file_path, is_bin, description in zip(file_paths, is_binary, descriptions):
            self.add_file(file_path, is_bin, description)
        self.on_change = on_change

    def get_all_file_widgets(self) -> List[FileWidget]:
        return self.file_widgets

    def add_file(self, abs_file_path='', is_binary=False, description=''):
        new_file_widget = FileWidget(self._abs_project_directory, len(self.file_widgets) + 1,
                                     file_path=abs_file_path, is_binary=is_binary, description=description,
                                     on_change=self.on_change)
        new_file_widget.browse_button.clicked.connect(self.on_browse)
        self.file_widgets.append(new_file_widget)
        new_file_widget.remove_button.clicked.connect(partial(self._remove_file, new_file_widget))

        self.update_file_count()
        self._update_file_labels(len(self.file_widgets) - 1)
        self.display_selected_file(len(self.file_widgets) - 1)
        self.on_change()

    def on_browse(self, file_widget: FileWidget):
        self._update_file_labels()

    def display_selected_file(self, index: int):
        if self.current_file_widget:
            self.file_display_area.removeWidget(self.current_file_widget)
            self.current_file_widget.hide()  # Hide the current widget
        if index < 0 or index >= len(self.file_widgets):
            self.current_file_widget = None
            return
        self.current_file_widget = self.get_all_file_widgets()[index]
        self.current_file_widget.show()
        self.file_display_area.addWidget(self.current_file_widget)

    def set_project_directory(self, abs_project_directory: Optional[Path]):
        self._abs_project_directory = abs_project_directory
        for file_widget in self.get_all_file_widgets():
            file_widget.set_project_directory(abs_project_directory)

    def _remove_file(self, widget):
        index = self.get_all_file_widgets().index(widget)
        # widget.destroyed.connect(self._update_file_labels)
        widget.deleteLater()
        self.file_widgets.remove(widget)

        self.update_file_count()
        index = min(index, len(self.file_widgets) - 1)
        self._update_file_labels(index)
        self.display_selected_file(index)
        self.on_change()

    def _update_file_labels(self, index: Optional[int] = None):
        """Update the labels of the files in the ComboBox."""
        for i, file_widget in enumerate(self.get_all_file_widgets()):
            file_widget.file_label_widget.setText(f"File #{i + 1}:")
        index = index if index is not None else self.file_selector.currentIndex()
        self.file_selector.blockSignals(True)
        self.file_selector.clear()
        for i, file_widget in enumerate(self.file_widgets):
            abs_file_path = file_widget.abs_path if file_widget.abs_path else ''
            file_name = self._get_file_display_name(i + 1, abs_file_path)
            self.file_selector.addItem(file_name)
        self.file_selector.setCurrentIndex(index)
        self.file_selector.blockSignals(False)

    def update_file_count(self):
        """Update the total number of files label."""
        self.file_count_label.setText(f"Total number of files: {len(self.file_widgets)}")

    def _get_file_display_name(self, file_number: int, abs_file_path: str) -> str:
        """Return the display name for the file in the format 'File #X: <file_stem>'."""
        s = f"File #{file_number}"
        if abs_file_path:
            file_stem = Path(abs_file_path).name  # Get the full file name (including extension)
            s += f": {file_stem}"
        return s

    def setDisabled(self, disable: bool):
        self.add_file_button.setDisabled(disable)
        for file_widget in self.get_all_file_widgets():
            file_widget.setDisabled(disable)

    def clear(self, num_files=1):
        """Clear all the file widgets and reset with a specified number of empty file slots."""
        # Clear all existing file widgets
        self.file_widgets.clear()
        self.file_selector.clear()

        # Remove widgets from the display area
        while self.file_display_area.count():
            widget = self.file_display_area.takeAt(0).widget()
            if widget:
                widget.deleteLater()
        for _ in range(num_files):
            self.add_file()

        # Update the file count label
        self.update_file_count()


class SingleFileWidget(MultiFileWidget):
    """
    Allows only one file.
    Do not show the 'Add Another File' button.
    Do not allow removing the file.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_file_button.hide()

    def add_file(self, abs_file_path='', is_binary=False, description=''):
        if self.files_layout.count() > 0:
            return
        super().add_file(abs_file_path, is_binary, description)
        self.get_all_file_widgets()[0].remove_button.hide()


class TextEditWithHeader(QWidget):
    def __init__(self, title: str, help_text: str = '', text: str = '', on_change=lambda: None):
        super().__init__()
        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(QLabel(title))
        text_edit = PlainTextPasteTextEdit()
        text_edit.setPlaceholderText(help_text)
        text_edit.setStyleSheet(text_box_style)
        text_edit.setPlainText(text)
        layout.addWidget(text_edit)
        self.text_edit = text_edit
        self.text_edit.textChanged.connect(on_change)  # Connect signal to callback

    def clear(self):
        self.text_edit.clear()

    def setPlainText(self, text):  # noqa for consistency with QTextEdit
        self.text_edit.setPlainText(text)

    def toPlainText(self):  # noqa for consistency with QTextEdit
        return self.text_edit.toPlainText()

    def setDisabled(self, disable: bool):
        self.text_edit.setReadOnly(disable)


class BaseStartDialog(QDialog):
    def __init__(self, steps_runner_cls: Type[BaseStepsRunner] = None,
                 project_directory: Optional[Path] = None):
        super().__init__()
        self.steps_runner_cls = steps_runner_cls
        self.current_config = {}
        self.is_locked = False
        self._abs_project_directory = None

        self._set_style()
        self.layout = QVBoxLayout(self)
        self.layout.addLayout(self._create_project_name_layout())
        self.widgets = self._create_widgets()
        for widget in self.widgets.values():
            self.layout.addWidget(widget)
        self.layout.addLayout(self._create_start_exist_buttons_layout())

        self._initialize(project_directory)

    def _create_widgets(self):
        return {}

    def _set_style(self):
        self.setWindowTitle(f"data-to-paper: Set and Run Project ({self.steps_runner_cls.name})")
        self.resize(1000, 1000)
        self.setStyleSheet(style_sheet)

    def _initialize(self, project_directory: Optional[Path]):
        self._clear_all()
        if project_directory:
            try:
                self.open_project(project_directory)
            except FileNotFoundError:
                self._clear_all()
                QMessageBox.warning(self, "Invalid Directory",
                                    f"The app was started with an invalid project directory:\n{project_directory}.\n"
                                    f"Please select a valid project directory.")

    def _create_project_name_layout(self):
        project_name_layout = QHBoxLayout()
        self.project_folder_header = QLabel("Project:")
        project_name_layout.addWidget(self.project_folder_header)
        self.project_folder_label = QLabel()
        self.project_folder_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        project_name_layout.addWidget(self.project_folder_label)
        self.open_button = QPushButton("Open")
        self.open_button.clicked.connect(lambda _: self.open_project())
        project_name_layout.addWidget(self.open_button)
        self.save_button = QPushButton("Save")
        self.save_button.clicked.connect(self.save_project)
        project_name_layout.addWidget(self.save_button)
        self.new_button = QPushButton("New")
        self.new_button.clicked.connect(self._clear_all)
        project_name_layout.addWidget(self.new_button)
        return project_name_layout

    def _create_start_exist_buttons_layout(self):
        buttons_tray = QHBoxLayout()

        start_button = QPushButton("Save and Start")
        start_button.clicked.connect(self.on_start_clicked)
        start_button.setDisabled(True)
        buttons_tray.addWidget(start_button)

        close_button = QPushButton("Exit")
        close_button.clicked.connect(self.close_app)
        buttons_tray.addWidget(close_button)

        return buttons_tray

    def _clear_all(self):
        self._set_project_directory(None)
        for widget in self.widgets.values():
            widget.clear()
        self._lock_project_for_editing(disable=False)
        self.current_config = {}

    def close_app(self):
        self.reject()
        QApplication.instance().exit()  # Exit the application

    def _lock_project_for_editing(self, disable=True):
        self.is_locked = disable
        self.project_folder_header.setText("Project [LOCKED]:" if disable else "Project:")
        self.save_button.setDisabled(disable)
        for widget in self.widgets.values():
            widget.setDisabled(disable)

    """Save project"""

    def _check_project(self) -> bool:
        # config = self._convert_widgets_to_config()
        return True

    def _browse_for_new_project_directory(self) -> Optional[Path]:
        dialog = QFileDialog(self, "Save Project: Select or create an empty directory",
                             str(BASE_PROJECT_DIRECTORY))
        dialog.setFileMode(QFileDialog.Directory)  # noqa
        dialog.setOption(QFileDialog.ShowDirsOnly, True)  # noqa
        dialog.setAcceptMode(QFileDialog.AcceptOpen)  # noqa
        if dialog.exec() != QFileDialog.Accepted:  # noqa
            return
        project_directory = Path(dialog.selectedFiles()[0])
        # check that the project directory is empty:
        if list(project_directory.iterdir()):
            QMessageBox.warning(self, "Invalid Directory",
                                "The selected directory is not empty. Please select an empty directory, "
                                "or create a new one.")
            return
        return project_directory

    def save_project(self) -> bool:
        if not self._check_project():
            return False
        if not self._abs_project_directory:
            project_directory = self._browse_for_new_project_directory()
            if not project_directory:
                return False
            self._set_project_directory(project_directory)
        self._save_project()
        return True

    def _save_project(self):
        if self.is_locked:
            return
        config = self._convert_widgets_to_config()
        self.steps_runner_cls.create_project_directory_from_project_parameters(
            self._abs_project_directory, config)

    def _check_data_files_exist(self) -> bool:
        config = self._convert_widgets_to_config()
        try:
            self.steps_runner_cls.check_files_exist(self._abs_project_directory, config)
        except FileNotFoundError as e:
            QMessageBox.warning(self, "Missing Files",
                                f"Cannot start the project:\n{e}")
            return False
        return True

    """Open project"""

    @property
    def project_file_name(self):
        return self.steps_runner_cls.PROJECT_PARAMETERS_FILENAME

    def open_project(self, project_directory: Optional[Path] = None):
        if project_directory is None:
            project_directory = self._browse_for_existing_project_directory()
            if not project_directory:
                return
        self._set_project_directory(project_directory)
        project_directory = self._abs_project_directory
        config = self.steps_runner_cls.get_project_parameters_from_project_directory(project_directory,
                                                                                     add_default_parameters=False)
        self.current_config = config
        self._convert_config_to_widgets(config)
        runs_folder = project_directory / 'runs'
        is_previous_runs = runs_folder.exists() and bool(list(runs_folder.iterdir()))
        self._lock_project_for_editing(disable=is_previous_runs)

    def _convert_config_to_widgets(self, config):
        pass

    def _browse_for_existing_project_directory(self) -> Optional[Path]:
        project_directory = QFileDialog.getOpenFileName(
            self, "Select a data file", str(BASE_PROJECT_DIRECTORY),
            f"{self.project_file_name} ({self.project_file_name});;All Files (*)")[0]

        if not project_directory:
            return  # user cancelled
        project_directory = Path(project_directory).parent

        # check that the project directory contains the project parameters file:
        if not (project_directory / self.project_file_name).exists():
            QMessageBox.warning(self, "Invalid Directory",
                                "The selected directory is not a valid project directory. "
                                f"Please select a directory, containing '{self.project_file_name}'.")
            return
        return project_directory

    def _set_project_directory(self, project_directory: Optional[Path]):
        self._abs_project_directory = project_directory
        if project_directory is None:
            self.project_folder_label.setText('Untitled')
        else:
            try:
                project_directory = project_directory.relative_to(BASE_PROJECT_DIRECTORY)
            except ValueError:
                pass
            self.project_folder_label.setText(str(project_directory))

    """Start project"""

    def _convert_widgets_to_config(self) -> dict:
        config = self.current_config
        return config

    def get_project_parameters(self):
        return self._abs_project_directory, self._convert_widgets_to_config()

    def on_start_clicked(self):
        if not self._check_project():
            return
        if not self.save_project():
            return
        if not self._check_data_files_exist():
            return
        self.accept()


class DataFilesStartDialog(BaseStartDialog):

    def _get_date_files_widget(self):
        return self.widgets['files_widget']

    def _create_widgets(self):
        return {
            'files_widget': MultiFileWidget(),
        }

    def _convert_config_to_widgets(self, config):
        files_widget = self._get_date_files_widget()
        files_widget.clear(num_files=0)
        for file_path, is_binary, description in zip(
                config.get('data_filenames', []),
                config.get('data_files_is_binary', []),
                config.get('data_file_descriptions', [])):
            files_widget.add_file(file_path, is_binary, description)
        super()._convert_config_to_widgets(config)

    def _convert_widgets_to_config(self) -> dict:
        config = super()._convert_widgets_to_config()
        data_filenames = []
        data_file_descriptions = []
        data_files_is_binary = []
        for file_widget in self._get_date_files_widget().get_all_file_widgets():
            if file_widget.abs_path:
                data_filenames.append(file_widget.abs_or_rel_path)
                data_file_descriptions.append(file_widget.description)
                data_files_is_binary.append(file_widget.is_binary)
        config['data_filenames'] = data_filenames
        config['data_files_is_binary'] = data_files_is_binary
        config['data_file_descriptions'] = data_file_descriptions
        return config

    def _set_project_directory(self, project_directory: Optional[Path]):
        super()._set_project_directory(project_directory)
        self._get_date_files_widget().set_project_directory(project_directory)

    def _get_mandatory_items_to_start(self):
        raise NotImplementedError

    def update_start_button_state(self):
        mandatory_items = self._get_mandatory_items_to_start()

        # access the start_button from the layout
        start_button = self.layout.itemAt(self.layout.count() - 1).itemAt(0).widget()
        start_button.setDisabled(any(not item for item in mandatory_items))

    def open_project(self, project_directory: Optional[Path] = None):
        super().open_project(project_directory)
        self.update_start_button_state()


def interactively_create_project_folder(steps_runner_cls: Type[BaseStepsRunner],
                                        project_directory: Optional[Path] = None) -> Tuple[Path, dict]:
    get_or_create_q_application_if_app_is_pyside()
    start_dialog = steps_runner_cls.APP_STARTUP_CLS(
        steps_runner_cls=steps_runner_cls,
        project_directory=project_directory)
    if start_dialog.exec() == QDialog.Accepted:  # noqa
        pass
    else:
        sys.exit(0)
    return start_dialog.get_project_parameters()



import pickle
import os
import hashlib
import time
import traceback

from traceback import FrameSummary

from contextlib import contextmanager
from dataclasses import asdict, dataclass

from pathlib import Path
from typing import Union

from data_to_paper.env import DELAY_CODE_RUN_CACHE_RETRIEVAL
from data_to_paper.utils.file_utils import run_in_directory
from data_to_paper.utils.print_to_file import print_and_log


def old_directory_hash(directory):
    """Create a hash based on all files in the directory."""
    hasher = hashlib.sha256()
    for path, dirs, filenames in os.walk(directory):
        for filename in sorted(filenames):
            file_path = os.path.join(path, filename)
            # Consider the file path and the file contents
            hasher.update(file_path.encode('utf-8'))
            with open(file_path, 'rb') as file:
                while chunk := file.read(8192):
                    hasher.update(chunk)
    return hasher.hexdigest()


def directory_hash(directory):
    """
    Create a hash based on all files in the directory, hashing only the relative path of each file
    from the specified directory.
    """
    hasher = hashlib.sha256()
    root_dir = os.path.abspath(directory)
    all_files = []

    # Walk the directory tree and capture all file paths
    for path, dirs, files in os.walk(root_dir):
        for file in files:
            full_path = os.path.join(path, file)
            relative_path = os.path.relpath(full_path, start=directory)  # Compute the relative path
            all_files.append((relative_path, full_path))  # Store both relative and full path

    # Sort all file paths by relative path
    all_files.sort(key=lambda x: x[0])

    # Hash each file's relative path and contents
    for relative_path, full_path in all_files:
        hasher.update(relative_path.encode('utf-8'))
        with open(full_path, 'rb') as file:
            while chunk := file.read(8192):
                hasher.update(chunk)

    return hasher.hexdigest()


def _read_file(filename):
    with open(filename, 'rb') as f:
        return f.read()


def _write_file(filename, content):
    with open(filename, 'wb') as f:
        f.write(content)


def _read_files(filenames):
    file_contents = {}
    for fname in filenames:
        file_contents[fname] = _read_file(fname)
    return file_contents


def _write_files(filenames):
    for fname, content in filenames.items():
        _write_file(fname, content)


# TODO: hack for python version compatibility. Remove when possible.
class CustomFrameSummary(FrameSummary):
    @property
    def end_lineno(self):
        # Return a default value or None if 'end_lineno' is not available
        return getattr(self, '_end_lineno', None)

    @end_lineno.setter
    def end_lineno(self, value):
        self._end_lineno = value


traceback.FrameSummary = CustomFrameSummary


@dataclass
class CacheRunToFile:
    """
    A class that caches the results of a 'run' method to a file.
    Also caches the files created during the run.
    Files pre-existing in the run directory are considered part of the run input.
    """
    cache_filepath: Union[str, Path] = None  # Path to the cache file, or None to disable caching

    def _get_instance_key(self) -> tuple:
        return tuple(asdict(self).values())

    def _get_run_directory_key(self) -> tuple:
        return (directory_hash(self._get_run_directory()), )

    def _get_run_directory_old_key(self) -> tuple:
        return (old_directory_hash(self._get_run_directory()), )

    def _get_run_directory(self):
        raise NotImplementedError

    def _run(self, *args, **kwargs):
        raise NotImplementedError

    def _load_cache(self, filename=None):
        filename = filename or self.cache_filepath
        if os.path.exists(filename):
            with open(filename, 'rb') as f:
                return pickle.load(f)
        return {}

    def _dump_cache(self, cache, filename=None):
        filename = filename or self.cache_filepath
        with open(filename, 'wb') as f:
            pickle.dump(cache, f)

    def run(self, *args, **kwargs):
        """
        Cache the results of a call to _run().
        """
        if self.cache_filepath is None:
            return self._run(*args, **kwargs)

        cache = self._load_cache()
        key = self._get_instance_key() + self._get_run_directory_key() \
            + tuple(args) + tuple(kwargs.items())

        old_key = self._get_instance_key() + self._get_run_directory_old_key() \
            + tuple(args) + tuple(kwargs.items())

        if old_key in cache:
            # replace old key with new key
            print(f"{self.__class__.__name__}: Replacing old key with new key.")
            cache[key] = cache.pop(old_key)
            self._dump_cache(cache)

        if key in cache:
            print_and_log(f"{self.__class__.__name__}: Using cached output.")
            time.sleep(DELAY_CODE_RUN_CACHE_RETRIEVAL.val)
            results, filenames = cache[key]
            with run_in_directory(self._get_run_directory()):
                _write_files(filenames)
            return results

        print_and_log(f"{self.__class__.__name__}: Running and caching output.")
        # Call the function and cache the result along with any created files
        with run_in_directory(self._get_run_directory()):
            with get_created_files() as created_files:
                results = self._run(*args, **kwargs)
            file_contents = _read_files(created_files)

        # Update cache
        cache[key] = (results, file_contents)
        self._dump_cache(cache)

        return results


def _listdir_files_only():
    return [f for f in os.listdir() if os.path.isfile(f)]


@contextmanager
def get_created_files():
    """
    Context manager for returning all new files created in the current directory.
    Files are returned as a sorted list of filenames.
    Note: The files are not deleted after the context manager exits.
    """
    # we need to preserve the filenames and metadata of the files
    preexisting_files_and_metadata = set()
    for filename in _listdir_files_only():
        preexisting_files_and_metadata.update(_get_filename_and_metadata(filename))

    created_files = []
    try:
        yield created_files
    finally:
        for filename in sorted(_listdir_files_only()):
            if _get_filename_and_metadata(filename) not in preexisting_files_and_metadata:
                created_files.append(filename)


def _get_filename_and_metadata(filename):
    return filename, os.stat(filename).st_mtime



"""
==================================================
| Main script for running data-to-paper projects |
==================================================

To run a project, use one of these syntaxes:

1. To interactively choose an existing project or create a new one, simply run:
    `python run.py`

2. To a run a project with one of the example datasets:
    `python run.py <project_name>`
    where <project_name> is one of:
        - default: Hypothesis testing - interactive setup (same as running without a project name)
        - diabetes: Hypothesis testing - diabetes dataset with an open research goal
        - social_network: Hypothesis testing - social network dataset with an open research goal
        - nrp_nicu: Hypothesis testing - NRP NICU dataset with a predefined research goal
        - ML_easy: ML Treatment Optimization dataset with a predefined research goal (easy)
        - ML_medium: ML Treatment Optimization dataset with a predefined research goal (medium)
        - ML_hard: ML Treatment Optimization dataset with a predefined research goal (difficult)
        - toy: Toy example of a simple two-step research process

3. Run a custom project:
    `python run.py --research_type <research_type> --project_folder <project_folder>`

    <research_type>:
    One of:
        - `hypothesis_testing`: Hypothesis testing (default if not provided)
        - `toy_example`: Toy example of a two-step research process
        [Additional research types will be added in the future]

    <project_folder>:
    The path to the project folder (containing the project data-to-paper-xxx.json file).
    The path can be absolute or relative to the `projects` of the repo (e.g., 'diabetes/diabetes_open_goal').
    If not provided, the project folder will be interactively chosen, or interactively created.

In all the above syntaxes, you can also provide a `--run_name` argument to specify the name of a folder,
within `runs` folder of the project folder, where the results will be saved.
Default is run_name is `run_001`.

The script will run the project and save the results in the specified run folder.

See the `env.py` file for setting other available options.
"""

import argparse
import os
import sys
import importlib.metadata

from data_to_paper.base_steps.run_all_steps import set_project_and_run
from data_to_paper.research_types.hypothesis_testing.steps_runner import HypothesisTestingStepsRunner
from data_to_paper.research_types.toy_example.steps_runner import ToyStepsRunner

# setup the QT_QPA_PLATFORM environment variable to 'xcb' if the system is Linux
if sys.platform == 'linux':
    os.environ['QT_QPA_PLATFORM'] = 'xcb'

# Currently supported research types:
# We currently only support simple hypothesis-testing research.
# A toy example of a two-step research process is also available
# to demonstrate how new research types can be added.
# We anticipate adding more research types in the future.
RESEARCH_TYPES_TO_STEPS_RUNNERS = {
    'hypothesis_testing': HypothesisTestingStepsRunner,
    'toy_example': ToyStepsRunner,
}

DEFAULT_RUN_NAME = 'run_001'

# Predefined projects:
RUN_PARAMETERS = {
    # Default settings: Hypothesis testing - interactive setup
    'default': [HypothesisTestingStepsRunner, None],

    # Specific demo datasets
    'diabetes': [HypothesisTestingStepsRunner, 'diabetes/open_goal'],
    'nrp_nicu': [HypothesisTestingStepsRunner, 'nrp_nicu/fixed_goal'],
    'ML_hard': [HypothesisTestingStepsRunner, 'ML/fixed_goal_hard'],
    'ML_medium': [HypothesisTestingStepsRunner, 'ML/fixed_goal_medium'],
    'ML_easy': [HypothesisTestingStepsRunner, 'ML/fixed_goal_easy'],
    'social_network': [HypothesisTestingStepsRunner, 'social_network/open_goal'],

    # Toy example of a two-step research process:
    'toy': [ToyStepsRunner, 'toy_example/prime_numbers'],
}


def get_version() -> str:
    try:
        return importlib.metadata.version('data-to-paper')
    except importlib.metadata.PackageNotFoundError:
        return "unknown"


def run():
    parser = argparse.ArgumentParser()

    parser.add_argument('project', type=str, nargs='?', default=None)
    parser.add_argument('--run_name', type=str, default=None)
    parser.add_argument('--project_folder', type=str, default=None)
    parser.add_argument('--research_type', type=str, default=None)
    parser.add_argument('--version', action='store_true', help="Display the version of the tool")
    args = parser.parse_args()
    if args.version:
        print(f"data-to-paper version {get_version()}")
        return

    project = args.project
    run_name = args.run_name
    project_folder = args.project_folder
    research_type = args.research_type

    run_name = run_name or DEFAULT_RUN_NAME

    if project:
        if project not in RUN_PARAMETERS:
            raise ValueError(f"Project '{project}' is not recognized.\n"
                             f"Please choose one of these pre-set projects {list(RUN_PARAMETERS.keys())}")
        if project_folder is not None:
            raise ValueError("You can't provide a project folder when using a predefined project")
        if research_type is not None:
            raise ValueError("You can't provide a research type when using a predefined project")
        steps_runner_cls, project_folder = RUN_PARAMETERS[project]
    else:
        research_type = research_type or 'hypothesis_testing'
        steps_runner_cls = RESEARCH_TYPES_TO_STEPS_RUNNERS[research_type]

    set_project_and_run(steps_runner_cls, project_directory=project_folder, run_name=run_name)


def main():
    run()


if __name__ == '__main__':
    main()



from typing import List, Optional

from data_to_paper.text.text_extractors import extract_text_between_tags


def get_displayitem_label(displayitem: str) -> Optional[str]:
    """
    Extract the label of a LaTeX display item (table, figure, etc.).
    """
    return extract_text_between_tags(displayitem, '\\label{', '}', keep_tags=False)


def get_displayitem_caption(displayitem: str, first_line_only: bool = False) -> Optional[str]:
    """
    Extract the caption of a floating LaTeX display item (table, figure, etc.).
    """
    caption = extract_text_between_tags(displayitem, '\\caption{', '}', keep_tags=False)
    if first_line_only:
        caption = caption.split('\n')[0]
    return caption


def add_displayitems_to_paper_section(section_content: str, displayitems: List[str]) -> str:
    """
    Insert the floating elements into the section content.
    """
    paragraphs = section_content.split('\n\n')
    displayitem_paragraph_numbers = []
    for displayitem in displayitems:
        displayitem_label = get_displayitem_label(displayitem)
        # find the paragraph that contains the displayitem reference
        for num, paragraph in enumerate(paragraphs):
            if displayitem_label is not None and displayitem_label in paragraph:
                # add the displayitem after the reference paragraph.
                displayitem_paragraph_numbers.append(num)
                break
        else:
            # add the displayitem at the end of the last paragraph.
            displayitem_paragraph_numbers.append(len(paragraphs) - 1)
    for displayitem_paragraph_number, displayitem in zip(displayitem_paragraph_numbers, displayitems):
        paragraphs[displayitem_paragraph_number] += '\n\n' + displayitem
    return '\n\n'.join(paragraphs)



from __future__ import annotations
import re
from dataclasses import dataclass, field
from typing import Optional, List, Tuple, Union

from data_to_paper.base_products.product import Product
from .file_view_params import ViewPurpose, ContentViewPurposeConverter
from .ref_numeric_values import numeric_values_in_products_pattern, ReferencedValue, HypertargetFormat, \
    HypertargetPosition


@dataclass
class ReferencableTextProduct(Product):
    referencable_text: BaseReferenceableText = None
    content_view_purpose_converter: Optional[ContentViewPurposeConverter] = \
        field(default_factory=ContentViewPurposeConverter)

    @property
    def hypertarget_prefix(self):
        return self.referencable_text.hypertarget_prefix

    def is_valid(self):
        return self.referencable_text is not None

    def _get_formatted_content(self, view_purpose: ViewPurpose) -> str:
        hypertarget_format = self.content_view_purpose_converter.convert_view_purpose_to_hypertarget_format(
            view_purpose)
        content, _ = self._get_formatted_text_and_header_references(hypertarget_format)
        return content

    def get_header(self, view_purpose: ViewPurpose = ViewPurpose.PRODUCT, **kwargs) -> str:
        hypertarget_format = self.content_view_purpose_converter.convert_view_purpose_to_hypertarget_format(
            view_purpose)
        _, references = self._get_formatted_text_and_header_references(hypertarget_format)
        header = super().get_header()
        header = '\n'.join(reference.to_str(hypertarget_format) for reference in references) + header
        return header

    def _get_content_as_formatted_text(self, level: int, view_purpose: ViewPurpose, **kwargs) -> str:
        content = self._get_formatted_content(view_purpose)
        content = self._process_content(content)
        return content

    def _process_content(self, content: str) -> str:
        return content

    def _get_formatted_text_and_header_references(self, hypertarget_format: HypertargetFormat
                                                  ) -> Tuple[str, List[ReferencedValue]]:
        return self.referencable_text.get_formatted_text_and_header_references(hypertarget_format)


def _num_to_letters(num: int) -> str:
    """
    Convert a number to letters.
    1 -> a, 2 -> b, 3 -> c, ...
    200 -> gv
    """
    letters = ''
    while num > 0:
        num, remainder = divmod(num - 1, 26)
        letters = chr(ord('a') + remainder) + letters
    return letters


def convert_str_to_latex_label(text: str, prefix: Optional[str] = None) -> str:
    """
    Convert str, like filename, into valid latex hypertarget label
    """
    if prefix is not None:
        text = f'{prefix}-{text}'
    return text.replace(".", "-").replace("_", "-")


@dataclass
class BaseReferenceableText:
    """
    A text which can be converted to hypertargeted text
    """
    hypertarget_prefix: Optional[str] = None  # if None, no hypertargets are created

    def get_formatted_text_and_references(self, hypertarget_format: HypertargetFormat
                                          ) -> Tuple[str, List[ReferencedValue]]:
        raise NotImplementedError

    def get_formatted_text(self, hypertarget_format: HypertargetFormat) -> str:
        return self.get_formatted_text_and_references(hypertarget_format)[0]

    def get_formatted_text_and_header_references(self, hypertarget_format: HypertargetFormat
                                                 ) -> Tuple[str, List[ReferencedValue]]:
        content, references = self.get_formatted_text_and_references(hypertarget_format)
        if hypertarget_format.is_hypertarget_position_header():
            return content, references
        return content, []


@dataclass
class FromTextReferenceableText(BaseReferenceableText):
    """
    A text which can be converted to hypertargeted text
    """
    text: str = ''
    pattern: str = r''

    def _get_reference(self, match, reference_num, line_no, line_no_with_ref, in_line_number):
        return ReferencedValue(
            label=self._get_label(match, reference_num, line_no, line_no_with_ref, in_line_number),
            value=self._get_value(match, reference_num, line_no, line_no_with_ref, in_line_number),
            is_target=True,
        )

    def _get_label(self, match, reference_num, line_no, line_no_with_ref, in_line_number):  # noqa
        in_line_letter = _num_to_letters(in_line_number + 1)
        return f'{self.hypertarget_prefix}{line_no_with_ref}{in_line_letter}'

    def _get_value(self, match, reference_num, line_no, line_no_with_ref, in_line_number):  # noqa
        return match.group(0)

    def get_formatted_text_and_references(self, hypertarget_format: HypertargetFormat
                                          ) -> Tuple[str, List[ReferencedValue]]:
        """
        Find all numeric values in text and create references to them.
        Replace the numeric values with the references.
        Return the text with the references and the references.
        """

        def replace_numeric_value_with_hypertarget(match):
            nonlocal reference_num, line_no, line_no_with_ref
            in_line_number = len(references_per_line)
            reference = self._get_reference(match, reference_num, line_no, line_no_with_ref, in_line_number)
            references_per_line.append(reference)
            reference_num = reference_num + 1
            if hypertarget_format and hypertarget_format.position != HypertargetPosition.HEADER:
                return reference.to_str(hypertarget_format)
            else:
                return reference.value

        if self.hypertarget_prefix is None and hypertarget_format:
            hypertarget_format = HypertargetFormat()
            # TODO: need to decide what to do here
            # raise ValueError("ReferenceableText has no hypertarget_prefix")
        reference_num = 0
        references = []
        lines = self.text.split('\n')
        line_no_with_ref = 0
        for line_no, line in enumerate(lines):
            references_per_line = []
            line = re.sub(self.pattern, replace_numeric_value_with_hypertarget, line)
            if references_per_line:
                line_no_with_ref += 1
                lines[line_no] = line
                references.extend(references_per_line)
        text = '\n'.join(lines)
        return text, references


@dataclass
class NumericReferenceableText(FromTextReferenceableText):
    """
    A text whose numeric values can be converted to hypertargets.
    """
    pattern: str = numeric_values_in_products_pattern


@dataclass
class ListReferenceableText(FromTextReferenceableText):
    """
    A text whose list items can be converted to hypertargets.
    """
    reference_list: List[ReferencedValue] = field(default_factory=list)

    def _get_reference(self, match, reference_num, line_no, line_no_with_ref, in_line_number):
        return self.reference_list[reference_num]


def hypertarget_if_referencable_text_product(text: Union[str, ReferencableTextProduct],
                                             view_purpose: ViewPurpose, **kwargs) -> str:
    """
    Create hypertargets if the text is a ReferencableTextProduct, otherwise return the text.
    """
    if isinstance(text, ReferencableTextProduct):
        return text.as_formatted_text(view_purpose=view_purpose, **kwargs)
    return text


@dataclass
class LabeledNumericReferenceableText(FromTextReferenceableText):
    """
    A text whose labeled numeric values can be converted to hypertargets.
    Numeric values should be labeled with @@<...>@@.
    """
    pattern: str = r'@@<(.+?)>@@'

    def _get_value(self, match, reference_num, line_no, line_no_with_ref, in_line_number):  # noqa
        return match.group(0)[3:-3]


def label_numeric_value(text: str) -> str:
    return '@@<' + text + '>@@'



import importlib

from data_to_paper.base_cast import Agent
from data_to_paper.base_cast.types import Profile

from .env import THEME_NAME

# load theme:
theme = importlib.import_module(f'data_to_paper.research_types.hypothesis_testing.themes.{THEME_NAME}')


class ScientificAgent(Agent):
    Performer = 'Performer'
    Director = 'Director'
    DataExplorer = 'DataExplorer'
    GoalReviewer = 'GoalReviewer'
    # DataPreprocessor = 'DataPreprocessor'
    PlanReviewer = 'PlanReviewer'
    Debugger = 'Debugger'
    InterpretationReviewer = 'InterpretationReviewer'
    Writer = 'Writer'
    CitationExpert = 'CitationExpert'
    # TableExpert = 'TableExpert'

    @classmethod
    def get_primary_agent(cls) -> Agent:
        return cls.Performer

    @classmethod
    def get_human_agent(cls) -> Agent:
        return cls.Director

    def get_conversation_name(self) -> str:
        return AGENTS_TO_CONVERSATION_NAMES[self]

    @property
    def profile(self) -> Profile:
        return getattr(theme, self.name)


assert all(agent.profile.agent_name == agent.name for agent in Agent), \
    f"Agent name in theme {THEME_NAME} does not match Agent enum"


AGENTS_TO_CONVERSATION_NAMES = {
    ScientificAgent.Performer: None,
    ScientificAgent.Director: 'get data',
    ScientificAgent.DataExplorer: 'explore data',
    # ScientificAgent.DataPreprocessor: 'preprocess data',
    ScientificAgent.GoalReviewer: 'review goal',
    ScientificAgent.PlanReviewer: 'review plan',
    ScientificAgent.Debugger: 'debug',
    ScientificAgent.InterpretationReviewer: 'interpretation',
    ScientificAgent.Writer: 'write paper',
    ScientificAgent.CitationExpert: 'add citations',
    # ScientificAgent.TableExpert: 'add tables',
}



import os
import pickle
import threading
import multiprocessing
import tempfile
import uuid

from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Tuple, Any

from data_to_paper.env import MAX_EXEC_TIME
from data_to_paper.utils.mutable import Mutable
from data_to_paper.run_gpt_code.code_runner import CodeRunner, is_serializable
from data_to_paper.utils.types import ListBasedSet

from .base_run_contexts import MultiRunContext
from .cache_runs import CacheRunToFile
from .exceptions import FailedRunningCode, CodeTimeoutException

# process.queue fails on Mac OS X with large objects. Use file-based transfer instead.
RUN_CACHE_FILEPATH = Mutable(None)

USE_THREADING = False


@dataclass
class CodeRunnerWrapper(CacheRunToFile):
    code: str = None  # code to run
    timeout_sec: int = MAX_EXEC_TIME.val
    code_runner: CodeRunner = field(default_factory=CodeRunner)
    run_in_separate_process: bool = True
    cache_filepath: Path = field(default_factory=lambda: RUN_CACHE_FILEPATH.val)  # None if not caching

    @property
    def run_folder(self):
        return self.code_runner.run_folder

    def run_code(self) -> Tuple[Any, ListBasedSet[str], MultiRunContext, Optional[FailedRunningCode]]:
        return self.code_runner.run(code=self.code)

    def _get_run_directory(self):
        return self.run_folder

    def _get_instance_key(self) -> tuple:
        return (self.code, )

    def run(self, *args, **kwargs) -> Tuple[Any, ListBasedSet[str], MultiRunContext, Optional[FailedRunningCode]]:
        return super().run(*args, **kwargs)

    def _run(self):
        if self.run_in_separate_process:
            return self.run_code_in_separate_process()
        return self.code_runner.run(code=self.code)

    def run_code_in_separate_process(self) \
            -> Tuple[Any, ListBasedSet[str], MultiRunContext, Optional[FailedRunningCode]]:
        """
        Run the provided code in a separate process and report exceptions or specific warnings.
        Calls `run_in_provided_process` which is a wrapper for `run`.
        """
        queue_or_filepath = f"subprocess_output_{uuid.uuid4()}_{os.getpid()}.pkl"
        queue_or_filepath = os.path.join(tempfile.gettempdir(), queue_or_filepath)
        if USE_THREADING:
            process_cls = threading.Thread
        else:
            process_cls = multiprocessing.Process
        try:
            process = process_cls(
                target=self._run_code_and_put_result_in_queue,
                args=(queue_or_filepath, ),
            )
        except (AttributeError, TypeError):
            for k, v in self.__dict__.items():
                if not is_serializable(v):
                    print(f'Attribute {k} is not serializable.')
            raise
        process.start()
        process.join(self.timeout_sec)
        if process.is_alive():
            if not USE_THREADING:
                process.terminate()  # Terminate the process if it's still alive after timeout
            process.join()
            result = (
                None,
                [],
                MultiRunContext(),
                FailedRunningCode(exception=CodeTimeoutException(self.timeout_sec))
            )
        else:
            with open(queue_or_filepath, 'rb') as f:
                result = pickle.load(f)
            os.remove(queue_or_filepath)
            if isinstance(result, Exception):
                raise result
        return result

    def _run_code_and_put_result_in_queue(self, queue_or_filepath):
        """
        Run the provided code and put the result in the queue.
        """
        code_runner = self.code_runner
        try:
            result = code_runner.run(code=self.code)
        except Exception as e:
            result = e
        with open(queue_or_filepath, 'wb') as f:
            pickle.dump(result, f)



from typing import Dict, Optional, Union, Type

import colorama

from data_to_paper.utils.print_to_file import print_and_log
from data_to_paper.conversation.stage import Stage

from .enum_types import PanelNames
from .human_actions import ButtonClickedHumanAction, TextSentHumanAction, HumanAction, RequestInfoHumanAction

REQUESTING_MISSING_TEXT = "Requesting..."


class BaseApp:
    """
    A base class for the application.
    Provides the main interface for the application, including:
    - show_text
    - edit_text
    """

    # make sure the app is singleton
    instance = None

    def __init__(self):
        # make sure the app is singleton
        if self.instance is not None:
            raise Exception("App is a singleton!")
        self.instance = self
        self.step_runner = None
        self.stage_to_reset_to = None
        self._panels_and_positions_to_headers = {(panel_name, position): '' for panel_name in PanelNames
                                                 for position in range(2)}

    @classmethod
    def get_instance(cls):
        if cls.instance is None:
            cls.instance = cls()
        return cls.instance

    def request_panel_continue(self, panel_name: PanelNames):
        pass

    def request_text(self, panel_name: PanelNames, initial_text: str = '',
                     title: Optional[str] = None,
                     instructions: Optional[str] = None,
                     optional_suggestions: Dict[str, str] = None) -> str:
        return initial_text

    def request_action(self, panel_name: PanelNames, initial_text: str = '',
                       title: Optional[str] = None,
                       instructions: Optional[str] = None,
                       in_field_instructions: Optional[str] = '',
                       optional_suggestions: Dict[str, str] = None) -> HumanAction:
        """
        Requests text from the user.
        User can choose to edit the text, or select one of the optional suggestions.
        """
        optional_suggestions = optional_suggestions or {}
        optional_suggestions = {"Initial": initial_text, **optional_suggestions}
        text = self.request_text(panel_name, initial_text, title, instructions, in_field_instructions,
                                 optional_suggestions)
        if text == REQUESTING_MISSING_TEXT:
            return RequestInfoHumanAction('AI')
        if text == initial_text:
            return ButtonClickedHumanAction('Initial')
        for suggestion_name, suggestion_content in optional_suggestions.items():
            if text == suggestion_content:
                return ButtonClickedHumanAction(suggestion_name)
        return TextSentHumanAction(text)

    def request_reset_to_step(self, step_name: str):
        pass

    def show_text(self, panel_name: PanelNames, text: str, is_html: bool = False,
                  scroll_to_bottom: bool = False):
        pass

    def set_focus_on_panel(self, panel_name: PanelNames):
        pass

    def advance_stage(self, stage: Union[Stage, int, bool]):
        """
        Advances the stage.
        stage:
            Stage: the stage to advance to.
            int: the index of the stage to advance to.
            True: advance to the end (all stages are completed).
            False: advance to the beginning (before the first stage).
        """
        pass

    def send_product_of_stage(self, stage: Stage, product_text: str):
        pass

    def initialize(self):
        pass

    def _run_all_steps(self):
        self.step_runner.run_all_steps()

    def _get_stages(self) -> Type[Stage]:
        if self.step_runner is None:
            return Stage
        return self.step_runner.stages

    def _set_status(self, panel_name: PanelNames, position: int, status: str = ''):
        pass

    def clear_stage_to_reset_to(self):
        self.stage_to_reset_to = None

    def set_status(self, panel_name: PanelNames, position: int, status: str = ''):
        self._panels_and_positions_to_headers[(panel_name, position)] = status
        self._set_status(panel_name, position, status)

    def get_status(self, panel_name: PanelNames, position: int) -> str:
        return self._panels_and_positions_to_headers[(panel_name, position)]

    def set_header(self, header: str):
        pass

    def send_api_usage_cost(self, stages_to_costs: Dict[Stage, float]):
        pass


class ConsoleApp(BaseApp):
    """
    A console-based application.
    Very simple implementation of the edit_text and show_text methods, for debugging purposes.
    """

    def initialize(self):
        self._run_all_steps()

    @staticmethod
    def get_multiline_input() -> str:
        lines = []
        while True:
            line = input()
            if not line:
                break
            lines.append(line)
        return '\n'.join(lines)

    def request_text(self, panel_name: PanelNames, initial_text: str = '',
                     title: Optional[str] = None,
                     instructions: Optional[str] = None,
                     optional_suggestions: Dict[str, str] = None) -> str:
        if panel_name != PanelNames.FEEDBACK:
            # In the console app, we only support the feedback panel:
            return initial_text
        color = colorama.Fore.BLUE
        light_color = colorama.Fore.LIGHTBLUE_EX
        print_and_log(title, color=light_color)
        print_and_log("Suggestions:", should_log=False, color=color)
        print_and_log(f"#{0}. {'Initial content'}:", should_log=False, color=light_color)
        print_and_log(initial_text or "<empty>", should_log=False, color=color)
        if optional_suggestions:
            for index, (suggestion_name, suggestion_content) in enumerate(optional_suggestions.items()):
                print_and_log(f"#{index + 1}. {suggestion_name}:", should_log=False, color=light_color)
                print_and_log(suggestion_content or "<empty>", should_log=False, color=color)
        print_and_log("Enter your text, or the number of the suggestion you want to use "
                      f"(0 - {len(optional_suggestions)}): (use Enter newline. End with an empty line to submit)",
                      should_log=False, color=color)
        text = self.get_multiline_input().strip()
        if text.isdigit():
            suggestion_index = int(text)
            if suggestion_index == 0:
                text = initial_text
            else:
                text = list(optional_suggestions.values())[suggestion_index - 1]
        print_and_log(f"Text received:", color=light_color)
        print_and_log(text or "<empty>", color=color)
        return text

    def show_text(self, panel_name: PanelNames, text: str, is_html: bool = False,
                  scroll_to_bottom: bool = False):
        # print_and_log(text)
        pass  # no need to print, user already sees all console messages.



from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class FormattedSection:
    label: Optional[str]
    section: str
    is_complete: bool = True

    def to_tuple(self):
        return self.label, self.section, self.is_complete

    def to_text(self):
        if not self.is_block:
            # regular text
            return self.section
        else:
            # block
            text = f'```{self.label}{self.section}'
            if self.is_complete:
                text += '```'
            return text

    @property
    def is_block(self):
        return self.label is not None


class FormattedSections(List[FormattedSection]):
    """
    Represents a text with multiple triple-backtick sections.

    We split the text into sections.
    For example:
    text =
    '''
    Hello here is my code:
    ```python
    a = 2
    ```
    is represnted by:
    [
    (None, "\nHello here is my code:\n", True),
    ('python', "\na = 2\n", True),
    ]
    """

    @classmethod
    def from_text(cls, text: str, strip_label: bool = True) -> FormattedSections:
        """
        Create FormattedSections from text.
        strip_label: if True, then ``` python \n etc``` is converted to label='python', not ' python '
        """
        sections = text.split('```')
        is_block = True
        self = cls()
        for i, section in enumerate(sections):
            is_block = not is_block
            if section == '':
                continue
            if is_block:
                is_single_line = '\n' not in section
                text_in_quote_line = section.split('\n')[0]
                if strip_label:
                    text_in_quote_line = text_in_quote_line.strip()
                if not is_single_line and (text_in_quote_line == '' or text_in_quote_line.isalpha()):
                    label = text_in_quote_line
                    section = '\n' + '\n'.join(section.split('\n')[1:])
                else:
                    label = ''
            else:
                label = None
            is_last = i == len(sections) - 1
            is_incomplete = is_last and is_block
            self.append(FormattedSection(label, section, not is_incomplete))
        return self

    def to_text(self) -> str:
        text = ''
        for i, formatted_section in enumerate(self):
            text += formatted_section.to_text()
        return text

    def get_first_block(self) -> Optional[FormattedSection]:
        for formatted_section in self:
            if formatted_section.is_block:
                return formatted_section
        return None

    def get_last_block(self) -> Optional[FormattedSection]:
        for formatted_section in reversed(self):
            if formatted_section.is_block:
                return formatted_section
        return None

    def get_all_blocks(self) -> List[FormattedSection]:
        return [formatted_section for formatted_section in self if formatted_section.is_block]

    def is_last_block_incomplete(self):
        last_block = self.get_last_block()
        return last_block is not None and not last_block.is_complete



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Any, TYPE_CHECKING, Dict

from data_to_paper.base_products.file_descriptions import DataFileDescriptions
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.code_and_output_files.output_file_requirements import OutputFileRequirementsToFileToContent
from data_to_paper.code_and_output_files.ref_numeric_values import HypertargetFormat, HypertargetPosition, \
    ReferencedValue
from data_to_paper.code_and_output_files.referencable_text import convert_str_to_latex_label
from data_to_paper.latex.clean_latex import wrap_as_latex_code_output, replace_special_latex_chars, \
    replace_non_utf8_chars
from data_to_paper.run_gpt_code.base_run_contexts import RunContext
from data_to_paper.text import wrap_as_block
from data_to_paper.text.highlighted_text import format_text_with_code_blocks

if TYPE_CHECKING:
    from data_to_paper.run_gpt_code.overrides.dataframes.dataframe_operations import DataframeOperations


@dataclass
class CodeAndOutput:
    name: str = None
    code: str = None
    result: Any = None
    created_files: \
        OutputFileRequirementsToFileToContent = field(default_factory=OutputFileRequirementsToFileToContent)
    code_name: str = None
    code_explanation: Optional[str] = None
    provided_code: Optional[str] = None
    contexts: Optional[Dict[str, RunContext]] = None
    dataframe_operations: Optional[DataframeOperations] = None
    description_of_created_files: DataFileDescriptions = None

    def _get_file_requirement_and_content(self, filename: str):
        return self.created_files.get_created_files_to_requirements_and_contents()[filename]

    def _get_file_requirement(self, filename: str):
        return self._get_file_requirement_and_content(filename)[0]

    def get_code_line_str_for_file(self, filename: str) -> Optional[str]:
        """
        Return a string which can be found in the line where we should go to when we want to see the code
        that created the file.
        """
        requirement, content = self._get_file_requirement_and_content(filename)
        return requirement.get_code_line_str_for_file(filename, content)

    def get_lineno_for_file(self, code: str, filename: str) -> Optional[int]:
        header = self.get_code_line_str_for_file(filename)
        if header is None:
            return None
        lines = code.split('\n')
        for i, line in enumerate(lines):
            if header in line:
                return i + 1
        return None

    @staticmethod
    def _add_hypertarget_to_code(code: str, label: str, lineno: int) -> str:
        """
        Add hypertarget to the code at the beginning of the line with the lineno
        """
        lines = code.split('\n')
        hypertarget_format = HypertargetFormat(position=HypertargetPosition.WRAP, raised=True, escaped=True)
        lines[lineno - 1] = ReferencedValue('', label).to_str(hypertarget_format) + lines[lineno - 1]
        return '\n'.join(lines)

    def _get_code_line_label_for_file(self, filename: str) -> str:
        requirement, content = self._get_file_requirement_and_content(filename)
        return requirement.get_code_line_label_for_file(filename, content)

    def _get_hypertarget_for_appendix_file_header(self, filename: str) -> str:
        return ReferencedValue(
            value='',
            label=convert_str_to_latex_label(filename, 'file'),
            is_target=True).to_str(HypertargetFormat(position=HypertargetPosition.WRAP))

    def _get_hyperlinked_header_for_appendix_file(self, filename: str) -> str:
        return ReferencedValue(
            value=replace_special_latex_chars(filename),
            label=self._get_hyperlink_label_for_file_header(filename),
            is_target=False).to_str(HypertargetFormat(position=HypertargetPosition.WRAP))

    def _get_hyperlink_label_for_file_header(self, filename: str) -> str:
        requirement, content = self._get_file_requirement_and_content(filename)
        return requirement.get_hyperlink_label_for_file_header(filename, content)

    def _get_code_with_hypertargets(self) -> str:
        code = self.code
        for filename in self.created_files.get_created_content_files():
            lineno = self.get_lineno_for_file(code, filename)
            if lineno is not None:
                code = self._add_hypertarget_to_code(
                    code, self._get_code_line_label_for_file(filename), lineno)
        return code

    def as_latex_for_appendix(self) -> str:
        s = f"\\section{{{self.name}}}\n"
        if self.code:
            s += "\\subsection{{Code}}\n"
            s += f"The {self.name} was carried out using the following custom code:\n"
            s += '\n\\begin{python}\n' + self._get_code_with_hypertargets() + '\n\\end{python}\n\n'
        if self.provided_code:
            s += f"\\subsection{{Provided Code}}\n"
            s += f"The code above is using the following provided functions:\n"
            s += '\n\\begin{python}\n' + self.provided_code + '\n\\end{python}\n\n'
        if self.code_explanation:
            s += "\\subsection{Code Description}\n"
            s += '\n' + self.code_explanation

        outputs = self.created_files.get_created_content_files_to_pretty_contents(
            view_purpose=ViewPurpose.FINAL_APPENDIX, header_level=None)
        if outputs:
            s += '\n\n' + "\\subsection{Code Output}"
            for filename, content in outputs.items():
                hypertarget = self._get_hypertarget_for_appendix_file_header(filename)
                hyperlinked_header = self._get_hyperlinked_header_for_appendix_file(filename)
                s += f'\n{hypertarget}\n\n\\subsubsection*{{{hyperlinked_header}}}'
                s += '\n\n' + wrap_as_latex_code_output(content)
        s = replace_non_utf8_chars(s)
        return s

    def to_text(self, with_header: bool = True):
        s = ''
        if with_header:
            if self.name is None:
                s = f"# Code and Output\n"
            else:
                s = f"# {self.name} Code and Output\n"
        if self.code:
            s += "## Code:\n"
            s += wrap_as_block(self.code, 'python') + '\n'
        if self.provided_code:
            s += "## Provided Code:\n"
            s += wrap_as_block(self.provided_code, 'python') + '\n'
        if self.code_explanation:
            s += "## Code Description:\n"
            s += wrap_as_block(self.code_explanation, 'latex') + '\n'
        if self.created_files:
            outputs = self.created_files.get_created_content_files_to_pretty_contents(
                view_purpose=ViewPurpose.APP_HTML, header_level=3)
        else:
            outputs = None

        if outputs:
            s += "## Code Output:\n"
            for filename, output in outputs.items():
                s += f"### {wrap_as_block(output, 'html')}\n"
        return s

    def as_html(self):
        return format_text_with_code_blocks(self.to_text(), from_md=True, is_html=True)



import multiprocessing
import os
import pickle
import tempfile
import uuid


def process_func(queue, func, *args, **kwargs):
    exception = None
    try:
        result = func(*args, **kwargs)
    except Exception as e:
        exception = e
        result = None
    result_exception = (result, exception)
    if isinstance(queue, str):
        with open(queue, 'wb') as f:
            pickle.dump(result_exception, f)
    else:
        queue.put(result_exception)


def run_func_in_separate_process(func, *args, in_separate_process=True,
                                 use_file_instead_of_queue=True,
                                 **kwargs):
    if not in_separate_process:
        return process_func(None, func, *args, **kwargs)
    if use_file_instead_of_queue:
        queue_or_filepath = f"subprocess_output_{uuid.uuid4()}_{os.getpid()}.pkl"
        queue_or_filepath = os.path.join(tempfile.gettempdir(), queue_or_filepath)
    else:
        queue_or_filepath = multiprocessing.Queue()
    process = multiprocessing.Process(target=process_func, args=(queue_or_filepath, func, *args), kwargs=kwargs)
    process.start()
    process.join()
    if use_file_instead_of_queue:
        with open(queue_or_filepath, 'rb') as f:
            result_exception = pickle.load(f)
        os.remove(queue_or_filepath)
    else:
        result_exception = queue_or_filepath.get()
    return result_exception



import re
from typing import List


def find_citation_ids(latex_string) -> List[str]:
    """
    Find all the citation ids in the latex string.

    For example, if the latex string is
    "one citation \\cite{citation_id1} two citations \\cite{citation_id2, citation_id3}", then the function will return
    ['citation_id1', 'citation_id2', 'citation_id3'].
    """
    pattern = r"\\cite{([^}]+)}"
    matches = re.findall(pattern, latex_string)
    citation_ids = [citation_id.strip() for match in matches for citation_id in match.split(',')]
    return citation_ids



# This module is not being used anymore. Timeout is now implemented by CodeRunnerWrapper
# The use of signal is really nice because it allows catching the specific line where the code timed out,
# but it only works on mac, and there is no such alternative for Windows.


import signal
import threading
import os
from dataclasses import dataclass
from typing import Type, Optional

from data_to_paper.run_gpt_code.base_run_contexts import RegisteredRunContext


def timeout_context(seconds, exception=TimeoutError):
    """
    Context manager to terminate a function if runtime is too long.
    """

    # return different context manager depending on the operating system
    if os.name == 'nt':
        return TimeoutWindowsContext(seconds=seconds, exception=exception)
    else:
        return TimeoutUnixContext(seconds=seconds, exception=exception)


@dataclass
class BaseTimeoutContext(RegisteredRunContext):
    seconds: int = 10
    exception: Type[Exception] = TimeoutError


@dataclass
class TimeoutUnixContext(BaseTimeoutContext):

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.signal_handler)
        signal.alarm(self.seconds)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)
        return super().__exit__(exc_type, exc_val, exc_tb)

    def signal_handler(self, signum, frame):
        raise self.exception(self.seconds)


@dataclass
class TimeoutWindowsContext(BaseTimeoutContext):
    stop_event: Optional[threading.Event] = None
    worker_thread: Optional[threading.Thread] = None

    def __enter__(self):
        self.stop_event = threading.Event()
        self.worker_thread = threading.Thread(target=self.target)
        self.worker_thread.start()
        self.worker_thread.join(timeout=self.seconds)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        is_stopped = self.stop_event.is_set()
        self.stop_event = None
        self.worker_thread = None
        if not is_stopped:
            raise self.exception(self.seconds)
        return super().__exit__(exc_type, exc_val, exc_tb)

    def target(self):
        try:
            yield
        finally:
            self.stop_event.set()



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from data_to_paper.types import HumanReviewType
from data_to_paper.conversation import ConversationManager, GeneralMessageDesignation, Message
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.utils.replacer import StrOrReplacer, format_value
from data_to_paper.utils.print_to_file import print_and_log_magenta
from data_to_paper.text.text_counting import is_bulleted_list
from data_to_paper.interactive import PanelNames
from data_to_paper.interactive.human_actions import RequestInfoHumanAction
from data_to_paper.interactive.human_review import HumanReviewAppInteractor
from data_to_paper.env import TEXT_WIDTH, PAUSE_AT_LLM_FEEDBACK, AUTO_TERMINATE_AI_REVIEW
from data_to_paper.run_gpt_code.code_utils import extract_content_of_triple_quote_block, FailedExtractingBlock, \
    IncompleteBlockFailedExtractingBlock
from data_to_paper.servers.model_engine import ModelEngine
from data_to_paper.conversation import Role

from .converser import Converser
from .result_converser import ResultConverser, Rewind


class CycleStatus(Enum):
    FAILED_CHECK_SELF_RESPONSE = 'failed_check_self_response'
    NOT_APPROVED_BY_OTHER = 'not_approved_by_other'
    APPROVED_BY_OTHER = 'approved_by_other'
    MAX_ROUNDS_EXCEEDED = 'max_rounds_exceeded'


@dataclass
class DualConverserGPT(Converser):
    """
    A base class for agents running two LLM agents.
    """
    COPY_ATTRIBUTES = {'other_conversation_name'}

    other_system_prompt: str = 'You are a helpful scientist.'

    other_conversation_name: str = None

    suppress_printing_other_conversation: bool = False

    def __post_init__(self):
        super().__post_init__()
        if self.other_conversation_name is None:
            self.other_conversation_name = f'{self.conversation_name} (other)'

        # For now, we do not allow the other conversation to continue a pre-existing conversation.
        assert self.other_conversation_name not in self.actions_and_conversations.conversations, \
            f'Conversation {self.other_conversation_name} already exists.'

        self.other_conversation_manager = ConversationManager(
            actions_and_conversations=self.actions_and_conversations,
            conversation_name=self.other_conversation_name,
            driver=self.driver if self.driver is not None else type(self).__name__,
            should_print=not self.suppress_printing_other_conversation,
        )

    @property
    def other_conversation(self):
        return self.other_conversation_manager.conversation

    def _pre_populate_other_background(self):
        """
        Add background messages to the other conversation.
        Only called if we are starting a new conversation.
        """
        pass

    def initialize_other_conversation_if_needed(self):
        self.other_conversation_manager.initialize_conversation_if_needed()
        if len(self.other_conversation) == 0:
            self.apply_to_other_append_system_message(self.other_system_prompt)
        self._pre_populate_other_background()

    def apply_to_other_get_and_append_assistant_message(self, tag: Optional[StrOrReplacer] = None,
                                                        comment: Optional[StrOrReplacer] = None,
                                                        is_code: bool = False, previous_code: Optional[str] = None,
                                                        model_engine: Optional[ModelEngine] = None,
                                                        hidden_messages: GeneralMessageDesignation = None,
                                                        expected_tokens_in_response: int = None,
                                                        **kwargs) -> Message:
        model_engine = model_engine or self.model_engine or ModelEngine.DEFAULT
        with self._app_temporarily_set_panel_status(PanelNames.FEEDBACK,
                                                    f'Waiting for Reviewer LLM ({model_engine})...'):
            self._app_send_prompt(PanelNames.FEEDBACK)
            message = self.other_conversation_manager.get_and_append_assistant_message(
                tag=tag,
                comment=comment,
                is_code=is_code, previous_code=previous_code,
                model_engine=model_engine,
                expected_tokens_in_response=expected_tokens_in_response,
                hidden_messages=hidden_messages, **kwargs)
            return message

    def apply_to_other_append_user_message(self, content: StrOrReplacer, tag: Optional[StrOrReplacer] = None,
                                           comment: Optional[StrOrReplacer] = None,
                                           ignore: bool = False,
                                           previous_code: Optional[str] = None, is_background: bool = False,
                                           send_to_app: Optional[bool] = None,
                                           app_panel: PanelNames = PanelNames.FEEDBACK,
                                           editing_title: str = None, editing_instructions: str = None,
                                           in_field_instructions: Optional[str] = None,
                                           sleep_for: Optional[float] = 0,
                                           **kwargs) -> Message:
        if send_to_app is None:
            send_to_app = not is_background and not ignore
        is_edit = editing_title or editing_instructions
        if is_edit:
            content = \
                self._show_and_edit_content(content, editing_title, editing_instructions, in_field_instructions,
                                            send_to_app, app_panel, sleep_for)
        result = self.other_conversation_manager.append_user_message(
            content=format_value(self, content),
            tag=tag,
            comment=comment,
            previous_code=previous_code,
            ignore=ignore, is_background=is_background, **kwargs)
        if not is_edit:
            self._show_and_edit_content(content, editing_title, editing_instructions, in_field_instructions,
                                        send_to_app, app_panel, sleep_for)
        return result

    def apply_to_other_append_system_message(self, content: StrOrReplacer, tag: Optional[StrOrReplacer] = None,
                                             comment: Optional[StrOrReplacer] = None,
                                             **kwargs) -> Message:
        return self.other_conversation_manager.append_system_message(
            content=format_value(self, content),
            tag=tag,
            comment=comment,
            **kwargs)

    def apply_to_other_append_surrogate_message(self, content: StrOrReplacer,
                                                tag: Optional[StrOrReplacer] = None,
                                                comment: Optional[StrOrReplacer] = None,
                                                ignore: bool = False,
                                                previous_code: Optional[str] = None,
                                                is_background: bool = False,
                                                **kwargs) -> Message:
        return self.other_conversation_manager.append_surrogate_message(
            content=format_value(self, content),
            tag=tag,
            comment=comment,
            previous_code=previous_code,
            ignore=ignore, is_background=is_background, **kwargs)

    def apply_to_other_delete_messages(self, message_designation: GeneralMessageDesignation,
                                       comment: Optional[str] = None):
        return self.other_conversation_manager.delete_messages(message_designation, comment=comment)


@dataclass
class DialogDualConverserGPT(DualConverserGPT, ResultConverser, HumanReviewAppInteractor):
    """
    A base class for agents running a dialog between two LLM agents (self and other), where the roles of the two
    agents are reversed. The ASSISTANT response from one conversation is used as the USER response in the other
    conversation, and vice versa.

    A conversation is formed between the two agents. The conversation is terminated when the other agent
    terminates the conversation by issuing a termination_phrase, or when a maximum number of rounds has been reached.

    The conversation can start with self either after its first response or before.
    """

    #                                               end if
    #                         can start             exceeds max_reviewing_rounds
    #                           here                  ^
    #                            |                    |
    #                            v               self_response
    #      end if             SELF.assistant ------------------->  OTHER.user
    # termination_phrase   <---  |                                     |
    # in other_response       SELF.user      <-------------------  OTHER.assistant
    #                           ^                other_response
    #                           |                round_num += 1
    #                         or, can
    #                         start here

    termination_phrase: str = 'Job completed'
    "A phrase used by the 'other' assistant to terminate the conversation."

    respond_to_ambiguous_reviewer_termination: str = None

    # TODO: Responding to ambiguous reviewer leads to the reviewer always apologizing and the conversation is not
    #  sensible anymore.
    #  This can only work if the reviewer is forced to return structured response, like triple-backtick block, or python
    #  list of strings, containing the feedback. or empty of for no feedback.

    # dedent_triple_quote_str("""
    #     Your answer is confusing because you have both provided feedback and included the phrase \t
    #     "{termination_phrase}".
    #     Please correct your response so that you EITHER include constructive feedback, OR just say \t
    #     "{termination_phrase}" without any other text.
    #     Do not apologize for your mistake/confusion - just provide the answer as is.
    #     """)

    append_termination_response_to_self: bool = True

    max_reviewing_rounds: int = 3
    max_reviewer_attempts: int = 4

    rewind_after_end_of_review: Optional[Rewind] = Rewind.AS_FRESH
    # can be
    # DELETE_ALL: delete the entire review including the user initiation prompt.
    # AS_FRESH: keep only the last response posted as fresh
    # ACCUMULATE (default, also None): keep all responses

    def __post_init__(self):
        if not self.are_we_reviewing_at_all:
            # we are not reviewing, so this is essentially a single conversation
            ResultConverser.__post_init__(self)
        else:
            super().__post_init__()
            # reverse roles:
            self.other_conversation_manager.assistant_agent = self.user_agent
            self.other_conversation_manager.user_agent = self.assistant_agent

        self.round_num = 0

    @property
    def are_we_reviewing_at_all(self) -> bool:
        return self.max_reviewing_rounds > 0 or self.actual_human_review

    @property
    def is_last_round(self) -> bool:
        return self.round_num >= self.max_reviewing_rounds

    def get_response_from_self_in_response_to_response_from_other(self, altered_other_response: str) -> Message:
        """
        Append response from other as user message to self conversation, and get response from assistant.
        """
        self.apply_append_user_message(altered_other_response,
                                       sleep_for=0 if self.actual_human_review else PAUSE_AT_LLM_FEEDBACK)
        return self.apply_get_and_append_assistant_message()

    def _alter_self_response(self, response: str) -> str:
        """
        Alter the response from self before sending it to other.
        """
        return response

    def _alter_other_response(self, response: str) -> str:
        """
        Alter the response from other before sending it to self.
        """
        return response

    def get_termination_phrase(self) -> str:
        return format_value(self, self.termination_phrase)

    def _is_reviewer_response_terminating(self, reviewer_response: str, is_human: bool = False) -> Optional[bool]:
        """
        Check if the response from the reviewer indicates that the reviewer is satisfied.
        True: the reviewer is satisfied and the dialog is completed.
        False: the reviewer is not satisfied and the dialog continues.
        None: the reviewer response is ambiguous.
        """
        is_phrase = self.get_termination_phrase().lower() in reviewer_response.lower()
        if is_human:
            return reviewer_response.strip() == '' or is_phrase
        if not is_phrase:
            return False
        if not is_bulleted_list(reviewer_response):
            return True
        return None

    def run_dialog(self) -> CycleStatus:
        """
        Run the dialog until it is completed.
        Returns the reason for termination.
        """
        conversation_len_before_first_round = len(self.conversation)
        while True:
            cycle_status = self.run_one_cycle()
            if cycle_status is not CycleStatus.NOT_APPROVED_BY_OTHER:
                break
        valid_result = self._get_valid_result()
        if self.rewind_after_end_of_review == Rewind.DELETE_ALL:
            self._rewind_conversation_to_first_response(-1, -1, start=conversation_len_before_first_round)
        elif self.rewind_after_end_of_review == Rewind.AS_FRESH:
            self._rewind_conversation_to_first_response(0, -1, start=conversation_len_before_first_round)
            self.apply_append_surrogate_message(self._convert_valid_results_to_fresh_looking_response(valid_result))
        return cycle_status

    def _convert_valid_results_to_fresh_looking_response_and_add_to_other(self) -> str:
        valid_result = self._get_valid_result()
        fresh_looking_self_response = self._convert_valid_results_to_fresh_looking_response(valid_result)
        altered_self_response = self._alter_self_response(fresh_looking_self_response)
        self.apply_to_other_append_user_message(altered_self_response)
        return altered_self_response

    def _get_other_message_and_response(self, auto_terminate: bool = True) -> (Message, str):
        should_terminate = auto_terminate and self.is_last_round
        self.round_num += 1
        if should_terminate:
            return None, self.get_termination_phrase()

        message = self.apply_to_other_get_and_append_assistant_message()
        if self.respond_to_ambiguous_reviewer_termination is not None:
            for attempt in range(self.max_reviewer_attempts):
                is_termination = self._is_reviewer_response_terminating(message.content)
                if is_termination is not None:
                    break
                # The reviewer response is ambiguous
                if attempt > 0:
                    self.apply_to_other_delete_messages(-1)  # delete the last message, to regenerate it
                else:
                    self.apply_to_other_append_user_message(self.respond_to_ambiguous_reviewer_termination)
                message = self.apply_to_other_get_and_append_assistant_message()

        return message, message.content

    def _add_or_replace_other_response(self, response: str):
        is_last_message_user = self.other_conversation[-1].role == Role.USER
        if not is_last_message_user:
            self.apply_to_other_delete_messages(-1)
            assert self.other_conversation[-1].role == Role.USER
        self.apply_to_other_append_surrogate_message(response)

    def _get_human_review(self, llm_review: Optional[str] = None, initial_review: Optional[str] = '') -> Optional[str]:
        goal = format_value(self, self.goal_noun)
        human_review, human_action = self._app_receive_text_and_action(
            PanelNames.FEEDBACK, initial_text=initial_review,
            title='User feedback requested',
            in_field_instructions=f'Give feedback on the {goal} (see Product tab above).\n'
                                  f'Leave blank if you have no suggestions for improvements.',
            optional_suggestions={'AI': llm_review, 'Default': ''})
        if isinstance(human_action, RequestInfoHumanAction):
            assert human_action.value == 'AI'
            return None
        return human_review

    def run_one_cycle(self) -> CycleStatus:
        """
        Run one cycle of the dialog. Makes updates to valid_result by calling
        _check_and_extract_value_from_self_response().
        """
        is_converged, is_new_valid_result = self._iterate_until_valid_response()
        if not is_new_valid_result:
            return CycleStatus.FAILED_CHECK_SELF_RESPONSE

        # We have a valid response from self. Now we can proceed with the dialog:
        if self.is_last_round and not self.actual_human_review:
            return CycleStatus.MAX_ROUNDS_EXCEEDED
        self._convert_valid_results_to_fresh_looking_response_and_add_to_other()

        if self.actual_human_review == HumanReviewType.NONE:
            llm_review_message, llm_review = self._get_other_message_and_response(auto_terminate=True)
            human_review = None
        elif self.actual_human_review == HumanReviewType.LLM_FIRST:
            llm_review_message, llm_review = self._get_other_message_and_response(AUTO_TERMINATE_AI_REVIEW)
            human_review = self._get_human_review(llm_review=llm_review)
        elif self.actual_human_review == HumanReviewType.LLM_UPON_REQUEST:
            llm_review_message, llm_review = None, None
            human_review = self._get_human_review()
            if human_review is None:
                llm_review_message, llm_review = self._get_other_message_and_response(AUTO_TERMINATE_AI_REVIEW)
                if self._is_reviewer_response_terminating(llm_review):
                    llm_review = self.get_termination_phrase()
                human_review = self._get_human_review(llm_review=llm_review, initial_review=llm_review)
        else:
            raise ValueError(f'Unknown HumanReviewType: {self.actual_human_review}')
        review = llm_review if human_review is None else human_review
        self._add_or_replace_other_response(review)
        if self._is_reviewer_response_terminating(review, is_human=human_review is not None):
            if self.append_termination_response_to_self:
                self.apply_append_user_message(
                    review, context=llm_review_message.context if llm_review_message else None,
                    sleep_for=not self.actual_human_review and PAUSE_AT_LLM_FEEDBACK.val)
            return CycleStatus.APPROVED_BY_OTHER

        altered_review = self._alter_other_response(review)
        self.get_response_from_self_in_response_to_response_from_other(altered_review)
        return CycleStatus.NOT_APPROVED_BY_OTHER


@dataclass
class ReviewDialogDualConverserGPT(DialogDualConverserGPT):
    """
    A base class for agents running a dialog between two LLM agents, where one is a "reviwee" who needs to perform
    a task towards a certain "goal", and the other is a "reviewer" who provides constructive feedback.

    The interaction proceeds in repeated cycles of the reviwee performing the task and the reviewer providing feedback.
    """

    # *** Properties that should be set according to the task we want to perform ***

    # roles:
    reviewer: str = 'scientific reviewer'

    other_system_prompt: str = dedent_triple_quote_str("""
        You are a {reviewer} for a {performer} who needs to {goal_verb} {goal_noun}.
        Your job is to advise me, the {performer}, and provide constructive bullet-point feedback in repeated cycles \t
        of improvements and feedback.

        When you feel that the goal has been achieved, respond explicitly with: 
        "{termination_phrase}" (approving-phrase)
        If you feel that the initial {goal_noun} is already good enough, it is perfectly fine and encouraged \t
        to respond with the approving-phrase immediately, without requesting any improvement cycles.
    """)

    sentence_to_add_at_the_end_of_reviewer_response: str = dedent_triple_quote_str("""\n\n
        Please correct your response according to any points you find relevant and applicable in my feedback.
        Send back a complete rewrite of the {goal_noun}.
        Make sure to send the full corrected {goal_noun}, not just the parts that were revised.
        """)

    sentence_to_add_at_the_end_of_performer_response: str = None

    def _alter_other_response(self, response: str) -> str:
        return response + '\n' + self.sentence_to_add_at_the_end_of_reviewer_response

    def _alter_self_response(self, response: str) -> str:
        response = super()._alter_self_response(response)
        if self.sentence_to_add_at_the_end_of_performer_response:
            return response + '\n' + self.sentence_to_add_at_the_end_of_performer_response
        else:
            return response

    def _print_conversation_header(self):
        print_and_log_magenta('==== Starting conversation ' + '=' * (TEXT_WIDTH - 27))
        print_and_log_magenta(self.conversation_name.center(TEXT_WIDTH))
        if self.are_we_reviewing_at_all:
            print_and_log_magenta(self.other_conversation_name.center(TEXT_WIDTH))
        print_and_log_magenta('=' * TEXT_WIDTH)

    def initialize_conversation_if_needed(self):
        super().initialize_conversation_if_needed()
        if self.are_we_reviewing_at_all:
            self.initialize_other_conversation_if_needed()

    def _run_and_return_termination_reason(self, with_review: bool = True) -> CycleStatus:
        if with_review:
            return self.run_dialog()
        else:
            return super()._run_and_return_termination_reason()


@dataclass
class QuotedReviewDialogDualConverserGPT(ReviewDialogDualConverserGPT):
    """
    A base class for agents running a dialog between two LLM agents, where one is a "reviwee" who needs to perform
    a task towards a certain "goal", and the other is a "reviewer" who provides constructive feedback.
    The performer is expected to return the goal as a triple-backtick block, so that it can be extracted.
    """

    your_response_should_be_formatted_as: str = 'a triple-backtick block (but send text, not code).'
    mission_prompt: str = ReviewDialogDualConverserGPT.mission_prompt \
        + '\nYour response should be formatted as {your_response_should_be_formatted_as}'

    sentence_to_add_at_the_end_of_reviewer_response: str = dedent_triple_quote_str("""\n\n
        Please correct your response according to any points you find relevant and applicable in my feedback.
        Send back a complete rewrite of the {goal_noun}.
        Remember, your response should be formatted as {your_response_should_be_formatted_as}
        """)

    rewind_after_getting_a_valid_response: Optional[Rewind] = Rewind.AS_FRESH

    def _convert_extracted_text_to_fresh_looking_response(self, extracted_text: str) -> str:
        return 'Here is the {goal_noun}:\n\n```' + extracted_text + '```\n\n'

    def _check_extracted_text_and_update_valid_result(self, extracted_text: str):
        self._check_flanked_response_is_not_just_header(extracted_text)
        self._update_valid_result(extracted_text)

    def _check_response_and_get_extracted_text(self, response: str) -> str:
        try:
            return extract_content_of_triple_quote_block(response, self.goal_noun, None)
        except FailedExtractingBlock as e:
            self._raise_self_response_error(
                title='# Failed to extract triple quote block',
                error_message=str(e),
                missing_end=isinstance(e, IncompleteBlockFailedExtractingBlock)
            )

    def _check_flanked_response_is_not_just_header(self, response: str):
        if response.count('\n') < 2 and response.count(' ') < 5:
            self._raise_self_response_error(
                title='# Response seems too short',
                error_message='Make sure you are flanking the entire response and not just the headers.'
            )



from __future__ import annotations
from typing import Optional, TYPE_CHECKING

from data_to_paper.run_gpt_code.run_issues import RunIssue, CodeProblem
from data_to_paper.run_gpt_code.user_script_name import is_called_from_user_script
from data_to_paper.utils.nice_list import NiceList

if TYPE_CHECKING:
    from data_to_paper.run_gpt_code.overrides.scipy.override_scipy import ScipyPValueOverride


def is_namedtuple(obj):
    return isinstance(obj, tuple) and hasattr(obj, '_fields')


class NoIterTuple:
    def __init__(self, _tuple, created_by: Optional[str] = None, context: Optional[ScipyPValueOverride] = None,
                 should_raise: bool = True, should_record: bool = True):
        self._tuple = _tuple
        self.created_by = created_by
        self.context = context
        self.should_raise = should_raise
        self.should_record = should_record

    def __getattr__(self, item):
        return getattr(self._tuple, item)

    def __repr__(self):
        return repr(self._tuple)

    def __str__(self):
        return str(self._tuple)

    def __len__(self):
        return len(self._tuple)

    def _get_fields(self):
        try:
            return self._tuple._fields
        except AttributeError:
            return None

    def _raise_or_register_if_called_from_user_script(self, exception: RunIssue):
        if not is_called_from_user_script(offset=4):
            return
        if self.should_record and self.context:
            self.context.issues.append(exception)
        if self.should_raise:
            raise exception

    def _get_instructions(self):
        return ('Your code should instead keep the results object and access its attributes by name when needed.\n'
                f'Attributes of `{self.created_by}` results object are: '
                f'{NiceList(self._get_fields(), wrap_with="`")}')

    def __getitem__(self, item):
        if isinstance(item, int):
            self._raise_or_register_if_called_from_user_script(RunIssue.from_current_tb(
                category='Coding: good practices',
                code_problem=CodeProblem.NonBreakingRuntimeIssue,
                issue=f'Accessing the results of {self.created_by} by index can lead to coding mistakes.',
                instructions=self._get_instructions()
            ))
        return getattr(self._tuple, item)

    def __iter__(self):
        self._raise_or_register_if_called_from_user_script(RunIssue.from_current_tb(
            category='Coding: good practices',
            code_problem=CodeProblem.NonBreakingRuntimeIssue,
            issue=f'Unpacking, or otherwise iterating over, the results of {self.created_by} '
                  f'can lead to coding mistakes.',
            instructions=self._get_instructions()
        ))
        return iter(self._tuple)



from typing import Dict

from data_to_paper.latex.clean_latex import replace_special_latex_chars
from data_to_paper.utils.dataframe import extract_df_axes_labels
from data_to_paper.text.text_formatting import escape_html


def convert_note_and_glossary_to_latex_table_caption(df, note: str, glossary: Dict[str, str], index: bool) -> str:
    """
    Convert a note and glossary to a latex string.
    """
    note_and_glossary = []
    if note:
        note_and_glossary.append(r'\item ' + replace_special_latex_chars(note))
    if glossary:
        axes_labels = extract_df_axes_labels(df, index=index)
        for key, value in glossary.items():
            if key in axes_labels or key == 'Significance':
                note_and_glossary.append(r'\item \textbf{' + replace_special_latex_chars(key) +
                                         '}: ' + replace_special_latex_chars(value))
    if len(note_and_glossary) == 0:
        note_and_glossary.append(r'\item ')  # add an empty item to avoid an error
    return '\n'.join(note_and_glossary)


def convert_note_and_glossary_to_latex_figure_caption(df, note: str, glossary: Dict[str, str], index: bool) -> str:
    note_and_glossary = []
    if note:
        note_and_glossary.append(replace_special_latex_chars(note))
    if glossary:
        axes_labels = extract_df_axes_labels(df, index=index)
        for key, value in glossary.items():
            if key in axes_labels or key == 'Significance':
                note_and_glossary.append((replace_special_latex_chars(key) +
                                          ': ' + replace_special_latex_chars(value)).strip())
    # add '. ' at the end of each line if missing:
    note_and_glossary = [line + ' ' if line.endswith('.') else line + '. ' for line in note_and_glossary]
    return '\n'.join(note_and_glossary)


def convert_note_and_glossary_to_html(df, note: str, glossary: Dict[str, str], index: bool) -> str:
    """
    Convert a note and glossary to an html string.
    """
    note_and_glossary = []
    if note:
        note_and_glossary.append(f'{escape_html(note)}<br>')
    if glossary:
        axes_labels = extract_df_axes_labels(df, index=index)
        for key, value in glossary.items():
            if key in axes_labels or key == 'Significance':
                note_and_glossary.append(f'<b>{escape_html(key)}</b>: {escape_html(value)}<br>')
    return '\n'.join(note_and_glossary)



from dataclasses import dataclass
from typing import Optional, Set, Iterable, Union, List

from data_to_paper.utils.print_to_file import print_and_log_red
from data_to_paper.base_cast import Agent
from data_to_paper.servers.llm_call import try_get_llm_response
from data_to_paper.servers.model_engine import OPENAI_CALL_PARAMETERS_NAMES, OpenaiCallParameters, ModelEngine
from data_to_paper.run_gpt_code.code_utils import add_label_to_first_triple_quotes_if_missing

from .actions_and_conversations import ActionsAndConversations, Conversations, Actions
from .conversation import Conversation
from .message import Message, Role, create_message, create_message_from_other_message
from .message_designation import GeneralMessageDesignation, convert_general_message_designation_to_list
from .conversation_actions import ConversationAction, AppendMessage, DeleteMessages, ResetToTag, \
    AppendLLMResponse, FailedLLMResponse, ReplaceLastMessage, \
    CreateConversation, AddParticipantsToConversation


@dataclass
class ConversationManager:
    """
    Manages an LLM conversation.
    Maintains a complete record of actions performed on the conversation.

    Allows processing Actions that act on the conversation.
    Maintains a list of these actions.
    """

    actions_and_conversations: ActionsAndConversations

    should_print: bool = True
    "Indicates whether to print added actions to the console."

    conversation_name: Optional[str] = None

    driver: str = ''
    "Name of the algorithm that is instructing this conversation manager."

    assistant_agent: Agent = None
    "The agent who is playing the assistant in the conversation."

    user_agent: Agent = None
    "The agent who is playing the user in the conversation."

    human_agent: Agent = None

    @property
    def conversations(self) -> Conversations:
        return self.actions_and_conversations.conversations

    @property
    def actions(self) -> Actions:
        return self.actions_and_conversations.actions

    @property
    def conversation(self) -> Conversation:
        return self.conversations.get_conversation(self.conversation_name)

    @property
    def participants(self) -> Set[Agent]:
        return {self.assistant_agent, self.user_agent}

    def _apply_action(self, action: ConversationAction):
        """
        Apply an action to a conversation and append to the actions list.
        """
        self.actions.apply_action(action, is_color=True)

    def _create_and_apply_action(self, action_type: type, **kwargs):
        """
        Create and apply an action to a conversation and append to the actions list.
        """
        self._apply_action(action_type(
            should_print=self.should_print,
            conversations=self.conversations,
            conversation_name=kwargs.pop('conversation_name', self.conversation_name),
            driver=kwargs.pop('driver', self.driver),
            **kwargs))

    def create_conversation(self):
        self._create_and_apply_action(CreateConversation, participants=self.participants)

    def add_participants(self, agents: Iterable[Agent]):
        self._create_and_apply_action(AddParticipantsToConversation, participants=set(agents))

    def initialize_conversation_if_needed(self):
        if self.conversation is None:
            self.create_conversation()
        else:
            if self.participants - self.conversation.participants:
                self.add_participants(self.participants - self.conversation.participants)

    def append_message(self, message: Message, comment: Optional[str] = None, **kwargs):
        """
        Append a message to a specified conversation.
        """
        self._create_and_apply_action(AppendMessage, message=message, comment=comment, **kwargs)

    def create_and_append_message(self, role: Role, content: str, tag: Optional[str], comment: Optional[str] = None,
                                  ignore: bool = False, previous_code: Optional[str] = None, is_code: bool = False,
                                  context: Optional[List[Message]] = None,
                                  is_background: bool = False, **kwargs):
        """
        Append a message to a specified conversation.
        """
        if role in [Role.ASSISTANT, Role.SURROGATE, Role.SYSTEM]:
            agent = self.assistant_agent
        elif role == Role.USER:
            agent = self.user_agent
        else:
            agent = None
        message = create_message(role=role, content=content, tag=tag, agent=agent, ignore=ignore,
                                 context=context, previous_code=previous_code, is_code=is_code,
                                 is_background=is_background)
        self.append_message(message, comment, **kwargs)

    def append_system_message(self, content: str, tag: Optional[str] = None, comment: Optional[str] = None,
                              ignore: bool = False,
                              is_background: bool = None, **kwargs):
        """
        Append a system-message to a specified conversation.
        """
        tag = tag or 'system_prompt'
        self.create_and_append_message(Role.SYSTEM, content, tag, comment,
                                       ignore=ignore, is_background=is_background,
                                       **kwargs)

    def append_user_message(self, content: str, tag: Optional[str] = None, comment: Optional[str] = None,
                            ignore: bool = False,
                            previous_code: Optional[str] = None, is_background: bool = False, **kwargs):
        """
        Append a user-message to a specified conversation.
        """
        self.create_and_append_message(Role.USER, content, tag, comment,
                                       ignore=ignore, previous_code=previous_code, is_background=is_background,
                                       **kwargs)

    def append_commenter_message(self, content: str, tag: Optional[str] = None,
                                 comment: Optional[str] = None, **kwargs):
        """
        Append a commenter-message to a specified conversation.

        Commenter messages are messages that are not sent to the LLM,
        rather they are just used as comments to the chat.
        """
        self.create_and_append_message(Role.COMMENTER, content, tag, comment, **kwargs)

    def append_surrogate_message(self, content: str, tag: Optional[str] = None, comment: Optional[str] = None,
                                 ignore: bool = False,
                                 previous_code: Optional[str] = None,
                                 is_background: bool = False, **kwargs):
        """
        Append a message with a pre-determined assistant content to a conversation (as if it came from the LLM).
        """
        self.create_and_append_message(Role.SURROGATE, content, tag, comment,
                                       ignore=ignore, previous_code=previous_code, is_background=is_background,
                                       **kwargs)

    def get_and_append_assistant_message(self, tag: Optional[str] = None, comment: Optional[str] = None,
                                         is_code: bool = False, previous_code: Optional[str] = None,
                                         is_json: bool = False,
                                         hidden_messages: GeneralMessageDesignation = None,
                                         expected_tokens_in_response: int = None,
                                         **kwargs  # for both create_message and openai params
                                         ) -> Message:
        """
        Get and append a response from openai to a specified conversation.

        If failed, retry while removing more messages upstream.
        """
        hidden_messages = convert_general_message_designation_to_list(hidden_messages)
        indices_and_messages = self.conversation.get_chosen_indices_and_messages(hidden_messages)
        actual_hidden_messages = hidden_messages.copy()

        # extract all OPENAI_CALL_PARAMETERS_NAMES from kwargs:
        openai_call_parameters = \
            OpenaiCallParameters(**{k: kwargs.pop(k) for k in OPENAI_CALL_PARAMETERS_NAMES if k in kwargs})

        # we try to get a response. if we fail we bump the model, and then gradually remove messages from the top,
        # starting at message 1 (we don't remove message 0, which is the system message).
        model = openai_call_parameters.model_engine or ModelEngine.DEFAULT
        while True:
            message = self._try_get_and_append_llm_response(tag=tag, comment=comment, is_code=is_code,
                                                            is_json=is_json,
                                                            previous_code=previous_code,
                                                            hidden_messages=actual_hidden_messages,
                                                            openai_call_parameters=openai_call_parameters,
                                                            expected_tokens_in_response=expected_tokens_in_response,
                                                            **kwargs)
            if isinstance(message, Message):
                return message

            # we failed to get a response. We start by bumping the model, if possible:
            try:
                model = model.get_model_with_more_context()
                model_was_bumped = True
            except ValueError:
                model_was_bumped = False
            if model_was_bumped:
                print_and_log_red(f'############# Bumping model #############')
                openai_call_parameters.model_engine = model
                continue

            # We have no option but to remove messages from the top:
            if len(indices_and_messages) <= 1:
                # we tried removing all messages and failed.
                raise RuntimeError('Failed accessing openai despite removing all messages from context.')
            print_and_log_red(f'############# Removing message from context #############')
            index, _ = indices_and_messages.pop(1)
            actual_hidden_messages.append(index)

    def _try_get_and_append_llm_response(self, tag: Optional[str], comment: Optional[str] = None,
                                         is_code: bool = False, previous_code: Optional[str] = None,
                                         is_json: bool = False,
                                         hidden_messages: GeneralMessageDesignation = None,
                                         openai_call_parameters: Optional[OpenaiCallParameters] = None,
                                         expected_tokens_in_response: int = None,
                                         **kwargs
                                         ) -> Union[Message, Exception]:
        """
        Try to get and append a response from openai to a specified conversation.

        The conversation is sent to openai after removing the messages with indices listed in hidden_messages.

        If getting a response is successful then append to the conversation, record action and return response string.
        If failed due to openai exception. Record a failed action and return the exception.
        """
        openai_call_parameters = openai_call_parameters or OpenaiCallParameters()
        if is_json:
            openai_call_parameters.response_format = {"type": "json_object"}
        messages = self.conversation.get_chosen_messages(hidden_messages)
        content = try_get_llm_response(messages, expected_tokens_in_response=expected_tokens_in_response,
                                       **openai_call_parameters.to_dict())
        if isinstance(content, Exception):
            self._create_and_apply_action(
                FailedLLMResponse, comment=comment, hidden_messages=hidden_messages, exception=content)
            return content

        if is_code:
            content = add_label_to_first_triple_quotes_if_missing(content, 'python')
        message = create_message(
            context=messages,
            role=Role.ASSISTANT, content=content, tag=tag, agent=self.assistant_agent,
            openai_call_parameters=None if openai_call_parameters.is_all_none() else openai_call_parameters,
            previous_code=previous_code, is_code=is_code,
            is_json=is_json)
        self._create_and_apply_action(
            AppendLLMResponse, comment=comment, hidden_messages=hidden_messages, message=message, **kwargs)
        return message

    def reset_back_to_tag(self, tag: str, comment: Optional[str] = None):
        """
        Reset the conversation to the last message with the specified tag.
        All messages following the message with the specified tag will be deleted.
        The message with the specified tag will be kept.
        """
        self._create_and_apply_action(ResetToTag, comment=comment, tag=tag)

    def delete_messages(self, message_designation: GeneralMessageDesignation, comment: Optional[str] = None):
        """
        Delete messages from a conversation.
        """
        self._create_and_apply_action(DeleteMessages, comment=comment, message_designation=message_designation)

    def replace_last_message(self, content: str, agent: Optional[Agent] = None, comment: Optional[str] = None
                             ) -> Message:
        """
        Replace the last message with the specified content.
        """
        message = create_message_from_other_message(
            other_message=self.conversation[-1], content=content, agent=agent)
        self._create_and_apply_action(ReplaceLastMessage, comment=comment, message=message)
        return message



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, List, Tuple

from data_to_paper.run_gpt_code.exceptions import FailedRunningCode
from data_to_paper.text import word_count
from data_to_paper.utils.replacer import format_value
from data_to_paper.utils.types import IndexOrderedEnum, ListBasedSet

MAX_WORDS_BEFORE_TERMINATING_ISSUE_LIST = 150


class CodeProblem(IndexOrderedEnum):
    """
    Code problems are sorted by severity, in the sense of progressively closer to a final fully working code.
    """
    NoCode = 'No code'
    IncompleteBlock = 'Incomplete block'
    NotSingleBlock = 'Not single block'
    StaticCheck = 'Static check'
    TimeoutError = 'Timeout error'
    RuntimeError = 'Runtime error'
    SyntaxError = 'Syntax error'
    MissingOutputFiles = 'Missing output files'
    NonBreakingRuntimeIssue = 'Non-breaking runtime issue'
    OutputFileCallingSyntax = 'Problem in the calling syntax of the output creation func'
    OutputFileContentA = 'Output file content first check (analysis stage)'
    OutputFileContinuity = 'Check dependency on previous output'
    OutputFileContentB = 'Output file content second check (displayitems stage)'
    OutputFileCompilation = 'Output file failed compilation'
    OutputFileAnnotation = 'Output file annotation'
    AllOK = 'All OK'

    def is_incomplete(self) -> bool:
        return self <= CodeProblem.IncompleteBlock

    def is_not_single_block(self) -> bool:
        return self == CodeProblem.NotSingleBlock

    def is_static_check(self) -> bool:
        return self == CodeProblem.StaticCheck

    def is_run_failed(self) -> bool:
        return CodeProblem.TimeoutError <= self <= CodeProblem.SyntaxError

    def is_missing_output_files(self) -> bool:
        return self == CodeProblem.MissingOutputFiles

    def is_run_completed_and_files_created(self) -> bool:
        return self >= CodeProblem.NonBreakingRuntimeIssue

    def get_stage(self) -> int:
        if self.is_incomplete():
            return 0
        elif self.is_not_single_block():
            return 1
        elif self.is_static_check():
            return 2
        elif self.is_run_failed():
            return 3
        elif self.is_missing_output_files():
            return 4
        elif self.is_run_completed_and_files_created():
            return 5
        else:
            raise NotImplementedError(f'Unknown problem stage for {self}')


@dataclass
class RunIssue(FailedRunningCode):
    code_problem: CodeProblem = None
    category: str = ''
    item: str = ''
    issue: str = ''
    instructions: str = ''
    comment: str = None
    end_with: Optional[str] = None
    requesting_small_change: bool = False
    forgive_after: Optional[int] = None  # Forgive after this many times,  None means never forgive

    @classmethod
    def from_current_tb(cls, code_problem: CodeProblem = None, category: str = None, item: str = None,
                        issue: str = None, instructions: str = None, comment: str = None,
                        end_with: Optional[str] = None,
                        requesting_small_change: bool = False, forgive_after: int = None, **kwargs):
        explicit_kwargs = {'code_problem': code_problem, 'category': category, 'item': item, 'issue': issue,
                           'instructions': instructions, 'comment': comment, 'end_with': end_with,
                           'requesting_small_change': requesting_small_change, 'forgive_after': forgive_after
                           }
        kwargs = kwargs | {k: v for k, v in explicit_kwargs.items() if v is not None}
        return super().from_current_tb(**kwargs)

    def formatted(self) -> RunIssue:
        return RunIssue(
            tb=self.tb,
            code_problem=self.code_problem,
            category=format_value(self, self.category),
            item=format_value(self, self.item),
            issue=format_value(self, self.issue),
            instructions=format_value(self, self.instructions),
            comment=format_value(self, self.comment),
            end_with=format_value(self, self.end_with),
            requesting_small_change=self.requesting_small_change,
            forgive_after=self.forgive_after,
        )

    def __str__(self):
        return RunIssues([self]).get_message_and_comment()[0]

    def __hash__(self):
        return hash((self.code_problem, self.category, self.item, self.issue, self.instructions, self.end_with))


class RunIssues(List[RunIssue]):

    def append_if_does_not_exist(self, issue: RunIssue):
        if issue not in self:
            self.append(issue)

    def get_message_and_comment(self, most_severe_only: bool = True, end_with: str = ''
                                ) -> Tuple[str, str, List[RunIssue]]:
        """
        We compose all the issues into a single message, and a single comment.
        """
        issues = [issue.formatted() for issue in self._get_issues(most_severe_only)]
        comments = ListBasedSet()

        if not issues:
            return 'All OK', '', []

        s = ''
        if len(issues) > 1:
            s += 'There are some issues that need to be corrected:\n\n'

        code_problems = sorted(set(issue.code_problem for issue in issues))
        posted_issues = []
        for code_problem in code_problems:
            categories = sorted(set(issue.category for issue in issues if issue.code_problem == code_problem))
            notes = []
            for category in categories:
                note = ''
                if category:
                    note += f'# {category}\n'
                issues_in_category = [issue for issue in issues if issue.category == category]
                unique_instructions = set(issue.instructions for issue in issues_in_category)
                shared_instructions = unique_instructions.pop() if len(unique_instructions) == 1 else None
                shared_instructions_word_count = word_count(shared_instructions) if shared_instructions else 0
                last_linenos_and_lines = None
                last_item = None
                for issue in issues_in_category:
                    posted_issues.append(issue)
                    if issue.item and issue.item != last_item:
                        last_item = issue.item
                        note += f'## {issue.item}:\n'
                    if issue.linenos_and_lines and issue.linenos_and_lines != last_linenos_and_lines:
                        last_linenos_and_lines = issue.linenos_and_lines
                        note += 'On line:\n'
                        note += '\n'.join(f'{lineno}: {line}' for lineno, line in issue.linenos_and_lines)
                        note += '\n'
                    note += f'{issue.issue.strip()}\n'
                    if shared_instructions is None and issue.instructions is not None:
                        note += f'{issue.instructions.strip()}\n'
                    note += '\n'
                    if issue.comment:
                        comments.add(issue.comment)
                    if word_count(note) + shared_instructions_word_count > MAX_WORDS_BEFORE_TERMINATING_ISSUE_LIST:
                        break
                if shared_instructions:
                    note += f'{shared_instructions}\n'
                notes.append(note)
            s += '\n\n'.join(notes)
        comment = '; '.join(comments)

        # Add the end_with message at the end:
        unique_end_with = set(issue.end_with for issue in issues)
        assert len(unique_end_with) == 1
        shared_end_with = unique_end_with.pop()
        if shared_end_with is not None:
            end_with = shared_end_with
        if end_with:
            s += f'\n{end_with}'
        return s, comment, posted_issues

    def get_most_severe_problem(self):
        return min(issue.code_problem for issue in self)

    def _get_issues(self, most_severe_only: bool = True) -> List[RunIssue]:
        if most_severe_only:
            return [issue for issue in self if issue.code_problem == self.get_most_severe_problem()]
        else:
            return list(self)

    def do_all_issues_request_small_change(self, highest_priority: bool = True) -> bool:
        return all(issue.requesting_small_change for issue in self._get_issues(highest_priority))
